MIN.PHI = 0.01 
MAX.PHI = 2

d1nb.mu = function(y,mu,phi) {
  # first derivative of dnbinom() on mu
  dnbinom(y,1/phi,mu=mu)*(y/mu-1)/(1+mu*phi)
  # for Poisson it is the same, so no need to speicify separately
  # dpois(y,mu)*(y/mu-1)
  # numerical version
  # grad(function(mu)  dnbinom(y,1/phi,mu=mu), mu)
}

d1nb.phi = function(y,mu,phi) {
  # first derivative of dnbinom() on phi
  r = 1/phi
  p = 1/(1+mu*phi)
  dnbinom(y,1/phi,mu=mu)*(-(digamma(y+r)-digamma(r)+log(p))*r^2+(y-mu)*p*r)
  # numerical version
  # grad(function(phi) dnbinom(y,1/phi,mu=mu), phi)
}

d2nb.mu = function(y,mu,phi) {
  # second derivative of dnbinom() on mu
  p = 1/(mu*phi+1)
  d1nb.mu(y,mu,phi)*p*(y/mu-1) - dnbinom(y,1/phi,mu=mu)*p^2*phi*(y/mu-1) - dnbinom(y,1/phi,mu=mu)*p*y/mu^2
  # numerical version
  # hessian(function(mu)  dnbinom(y,1/phi,mu=mu), mu)
}

d2nb.phi = function(y,mu,phi) {
  # 2nd derivative of dnbinom() on phi
  r = 1/phi
  p = 1/(1+mu*phi)
  A = -(digamma(y+r)-digamma(r)+log(p))*r^2+(y-mu)*p*r
  dA = ((psigamma(y+r,1)-psigamma(r,1))*r^2+mu*p)*r^2 + 2*(digamma(y+r)-digamma(r)+log(p))*r^3 - (y-mu)*p*r*(mu*p+r)
  d1nb.phi(y,mu,phi)*A + dnbinom(y,1/phi,mu=mu)*dA
  # numerical version
  # hessian(function(phi) dnbinom(y,1/phi,mu=mu), phi)
}

d2nb.mu_phi = function(y,mu,phi) {
  # partial derivative of dnbinom() on mu and phi
  r = 1/phi
  p = 1/(1+mu*phi)
  (d1nb.phi(y,mu,phi)*p-dnbinom(y,1/phi,mu=mu)*mu*p^2)*(y/mu-1)
  # numerical version
  # grad(function(phi) grad(function(mu) dnbinom(y,1/phi,mu=mu), mu), phi)
}

dmix = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # mixture density of count (y) given singal mean (mu), signal dispersion (phi), background mean (mu.bg), background dispersion (phi.bg).
  x = 0:y
  res = sum(dnbinom(x,1/phi,mu=mu)*dnbinom(y-x,1/phi.bg,mu=mu.bg))
  res
})

d1mix.mu = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # first derivative of dmix() on mu
  x = 0:y
  sum(d1nb.mu(x,mu,phi)*dnbinom(y-x,1/phi.bg,mu=mu.bg))
  # numerical version
  # grad(function(mu) dmix(y,mu,phi,mu.bg,phi.bg), mu)
})

d1mix.phi = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # first derivative of dmix() on phi
  x = 0:y
  sum(d1nb.phi(x,mu,phi)*dnbinom(y-x,1/phi.bg,mu=mu.bg))
  # numerical version
  # grad(function(phi) dmix(y,mu,phi,mu.bg,phi.bg), phi)
})

d2mix.mu = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # second derivative of dmix() on mu
  x = 0:y
  sum(d2nb.mu(x,mu,phi)*dnbinom(y-x,1/phi.bg,mu=mu.bg))
  # numerical version
  # hessian(function(mu) dmix(y,mu,phi,mu.bg,phi.bg), mu)
})

d2mix.phi = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # second derivative of dmix() on phi
  x = 0:y
  sum(d2nb.phi(x,mu,phi)*dnbinom(y-x,1/phi.bg,mu=mu.bg))
  # numerical version
  # hessian(function(phi) dmix(y,mu,phi,mu.bg,phi.bg), phi)
})

d2mix.mu_phi = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # partial derivative of dmix() on mu and phi
  x = 0:y
  sum(d2nb.mu_phi(x,mu,phi)*dnbinom(y-x,1/phi.bg,mu=mu.bg))
  # numerical version
  # grad(function(phi) grad(function(mu) dmix(y,mu,phi,mu.bg,phi.bg), mu), phi)
})

d1loglik.par = function(y,x,par,mu.bg,phi.bg) {
  beta = as.matrix(par[-length(par)])
  phi = par[length(par)]
  mu = exp(as.vector(x%*%beta))
  d = dmix(y,mu,phi,mu.bg,phi.bg)
  dbeta = t(d1mix.mu(y,mu,phi,mu.bg,phi.bg)/d*mu)%*%x
  dphi  = sum(d1mix.phi(y,mu,phi,mu.bg,phi.bg)/d)
  score = c(dbeta,dphi)
  if(length(names(par))==length(par)) par.names = names(par)
  else par.names = c(colnames(x),"phi")
  names(score) = par.names
  score
  # numerical version
  # grad(function(par) loglik(y,x,par,mu.bg,phi.bg), par)
}

d2loglik.par = function(y,x,par,mu.bg,phi.bg) {
  len = length(par)
  beta = as.matrix(par[-len])
  phi = par[len]
  mu = exp(as.vector(x%*%beta))
  d = dmix(y,mu,phi,mu.bg,phi.bg)
  dmu = d1mix.mu(y,mu,phi,mu.bg,phi.bg)
  dphi = d1mix.phi(y,mu,phi,mu.bg,phi.bg)
  C = dmu/d*mu
  d2beta = t(x)%*%diag(as.vector(d2mix.mu(y,mu,phi,mu.bg,phi.bg)/d*mu^2 - C^2 + C),length(y))%*%x
  d2phi = sum(d2mix.phi(y,mu,phi,mu.bg,phi.bg)/d-dphi^2/d^2)
  d2beta_phi = t((d2mix.mu_phi(y,mu,phi,mu.bg,phi.bg)/d - dmu*dphi/d^2)*mu)%*%x
  mat = matrix(NA,len,len)
  mat[-len,-len] = d2beta
  mat[-len,len] = mat[len,-len] = d2beta_phi
  mat[len,len] = d2phi
  mat
  if(length(names(par))==length(par)) par.names = names(par)
  else par.names = c(colnames(x),"phi")
  rownames(mat) = colnames(mat) = par.names
  mat
  # numerical version
  # hessian(function(par) loglik(y,x,par,mu.bg,phi.bg), par)
}

dgbinom = function(y,mu,phi=NA) {
  if(phi==0 | invalid(phi)) dpois(y,mu)
  else if(phi>0) dnbinom(y,1/phi,mu=mu)
  else if(phi<0 & phi>-1/mu) dbinom(y,-1/phi,-mu*phi)
  else 0
}

dgmix = Vectorize(function(y,mu,phi,mu.bg,phi.bg) {
  # mixture density of count (y) given singal mean (mu), signal dispersion (phi), background mean (mu.bg), background dispersion (phi.bg).
  x = 0:y
  res = sum(dgbinom(x,1/phi,mu=mu)*dgbinom(y-x,1/phi.bg,mu=mu.bg))
  res
})

loglik = function(y,x,par,mu.bg,phi.bg) {
  beta = as.matrix(par[-length(par)])
  phi = par[length(par)]
  mu = exp(as.vector(x%*%beta))
  d = dgmix(y,mu,phi,mu.bg,phi.bg)
  if(any(is.na(d) | d==0)) return(-Inf)
    #return(length(d)*log(.Machine$double.xmin))
  return(sum(log(d)))
}

lm.func = function(y,x,mu.bg,scaling,method) {
  # Check for groups with one observation
  y = (y - mu.bg)/scaling
  y[y<1] = 1
  fit = NULL
  phi.est = NA
  if(method=="lm.nb") {
    temp.df1 = data.frame(y=round(y),x,check.names=F)
    fit = tryCatch(glm.nb(y~0+.,data=temp.df1),
                   error = function(e) return(NULL))
  }
  if(is.null(fit) || !fit$converged || fit$theta>1/MIN.PHI || fit$theta<1/MAX.PHI) {
    cat("lm.nb failed. Use log.linear instead.\n")
    method = "loglinear"
  } else {
    loglik = fit$twologlik/2
    phi.est = 1/fit$theta
  }
  if(method=="loglinear") {
    y[y<1] = 1
    temp.df2 = data.frame(y = log(y),x,check.names=F)
    fit = lm(y~0+.,data=temp.df2)
    loglik = as.numeric(logLik(fit))
  }
  coefficients = summary(fit)$coefficients
  if (method=="lm.nb") {
    coefficients[,4] = 2*pt(-abs(coefficients[,3]),fit$df.residual)
  }
  if(nrow(coefficients) < ncol(x)) {
    failed.run.message.a <- "\nPossible Solutions: There are too many parameters to estimate given the number of samples. (1) Select fewer covariates for DE; (2) combine/reduce levels for categorical covariates; or (3) increase the number of samples. Afterwards re-run the Advanced Analysis."
    # cat(failed.run.message.a, file=log, sep='\n', append=TRUE)
    stop(failed.run.message.a)
  }
  # coefficients = coefficients[match(predictors,rownames(coefficients)),,drop=F]
  colnames(coefficients) = c("Estimate","Std.Error","Test statistic","pvalue")
  rownames(coefficients) = gsub("^`|`$","",rownames(coefficients))
  list(coefficients = coefficients,
       phi = unname(phi.est),
       df = fit$df.residual,
       loglik = loglik,
       method = method)
}

mle.func = function(y,x,phi=NA,mu.bg,phi.bg,scaling,fixed.args=list()) {
  # estimate parameter by MLE
  # initial estimate from lm.nb
  init.est = lm.func(y,x,mu.bg,scaling,"lm.nb")
  init = init.est$coefficients[,"Estimate"]
  names(init) = rownames(init.est$coefficients)
  init = c(init,phi=MIN.PHI)
  if(init.est$method=="lm.nb") init["phi"] = init.est$phi
  if(!invalid(phi)) {
    init["phi"] = phi
    fixed.args$phi = phi
  }
  init.par = as.list(init)
  
  y.scaled = round((y-mu.bg)/scaling+mu.bg) 
  y.scaled[y.scaled<1] = 1
  minuslogl = eval(parse(text=paste(sprintf("function(%s) {
    par = unlist(as.list(environment()))
    -loglik(y.scaled,x,par,mu.bg,phi.bg)
  }",paste(names(init),collapse=",")))))
  
  est = tryCatch({
    mle(minuslogl,start=init.par,fixed=fixed.args,nobs=length(y),method="L-BFGS-B")
  }, error = function(err) {
    print(err)
    return(NULL)
  })
  # if it fails, fix phi
  if(is.null(est) || coef(est)["phi"]>MAX.PHI) {
    init.par$phi = NULL
    fixed.args = list(phi=ifelse(is.na(init.est$phi), MIN.PHI, init.est$phi))
    cat("Estimation failed. Use fixed dispersion parameter instead.\n")
    est = tryCatch({
      mle(minuslogl,start=init.par,fixed=fixed.args,nobs=length(y),method="L-BFGS-B")
    }, error = function(err) {
      print(err)
      return(NULL)
    })
  }
  if(is.null(est)) return(NULL)
  coefficients = slot(summary(est),"coef")
  coefficients = coefficients[rownames(coefficients)!="phi",,drop=F]
  phi.est = attr(est,"fullcoef")["phi"]
  par.est = list(coefficients = coefficients, 
                 phi = unname(phi.est),
                 df = init.est$df,
                 loglik = -attr(est,"min"), 
                 vcov = attr(est,"vcov"),
                 hessian = slot(est,"details")$hessian,
                 converged = slot(est,"details")$convergence==0, 
                 method = attr(est,"method"))
  par.est
}

test.func = function(y,x,phi=NA,mu.bg,phi.bg,scaling,predictors,method) {
  # wrapper of the three methods
  y = as.vector(y)
  if(ncol(x)<=1 | length(predictors)==0) stop("No predictors specified.")
  if(method %in% c("lm.nb","loglinear")) return(lm.func(y,x,mu.bg,scaling,method))
  coefficients = tryCatch({
    test.mle(y,x,phi,mu.bg,phi.bg,predictors,scaling,method)
  }, error = function(e) {
    lm.func(y,x,mu.bg,scaling,"lm.nb")
  })
  coefficients
}

test.mle = function(y,x,phi,mu.bg,phi.bg,predictors,scaling,method) {
  # Wald, LR and Score test
  # Wald variance under the full model
  # Score variance under the reduced model

  # Estimate under the full model
  y.scaled = round(y/scaling)
  est.full = mle.func(y,x,phi=NA,mu.bg,phi.bg,scaling)
  if(is.null(est.full) || !est.full$converged || any(diag(est.full$vcov)<0)) {
    cat("Estimation under the full model did not converge. Use lm.nb instead.\n")
    return(lm.func(y,x,mu.bg,scaling,"lm.nb"))
  }
  par.len = nrow(est.full$coefficients) + 1
  par.full = c(est.full$coefficients[,"Estimate"],phi=est.full$phi)
  var.full = est.full$vcov
  loglik.full = est.full$loglik

  # Estimate under the reduced model
  coefficients = NULL
  for(predictor in predictors) {
    Estimate = par.full[predictor]
    var.predictor = var.full[predictor,predictor]
    # what if var<0
    sd = sqrt(var.predictor)
    if(method=="Wald") {
      test.stat = Estimate/sd
      pvalue = 2*pt(-abs(test.stat),est.full$df-1)
      stat.name = "t value"
    } else {
      est.reduced = eval(parse(text=sprintf("mle.func(y,x,phi=NA,mu.bg,phi.bg,scaling,fixed.args=list(%s=0))",predictor)))
      if(is.null(est.reduced) || !est.reduced$converged || 
         any(diag(est.reduced$vcov)<0) || loglik.full < est.reduced$loglik) {
        cat("Estimation under the reduced model did not converge. Use lm.nb instead.\n")
        return(lm.func(y,x,mu.bg,scaling,"lm.nb"))
      }
    }
    if(method=="LR") {
      test.stat = 2*(loglik.full - est.reduced$loglik)
      pvalue = 1 - pchisq(test.stat,1)
      # stat.name = "ChiSq value"
    } 
    if(grepl("Score",method)) {
      idx = which(rownames(est.full$coefficients)==predictor)
      par.reduced = c(est.reduced$coefficients[,"Estimate"],phi=est.reduced$phi)
      par.reduced = c(par.reduced[1:(idx-1)],0,par.reduced[idx:length(par.reduced)])
      names(par.reduced) = names(par.full)
      # score = d1loglik.par(y,x,par.reduced,mu.bg,phi.bg)[predictor]
      gradient = grad(function(par) loglik(y.scaled,x,par,mu.bg,phi.bg), par.reduced)
      names(gradient) = names(par.full)
      score = gradient[predictor]

      # fisher.reduced = -d2loglik.par(y.scaled,x,par.reduced,mu.bg,phi.bg)[-par.len,-par.len]
      fisher.reduced = -hessian(function(par) loglik(y.scaled,x,par,mu.bg,phi.bg), par.reduced)
      rownames(fisher.reduced) = colnames(fisher.reduced) = names(par.full)
      var.reduced = tryCatch(solve(fisher.reduced),
                             error = function(err) { return(NULL) } )
      if(is.null(var.reduced) || any(diag(var.reduced)<0)) {
        cat("Estimation under the reduced model did not converge. Use lm.nb instead.\n")
        return(lm.func(y,x,mu.bg,scaling,"lm.nb"))
      }
      test.stat = score*sqrt(var.reduced[predictor,predictor])
      # score1 uses z test, score2 uses t test
      if(method=="Score2") {
        pvalue = 2*pt(-abs(test.stat),est.full$df-1)
        # stat.name = "t value"
      }
      if(method == "Score1") {
        pvalue = 2*pnorm(-abs(test.stat))
        # stat.name = "z value"
      }
    }
    coef = unname(c(Estimate,sd,test.stat,pvalue))
    coefficients = rbind(coefficients,coef)
  }
  rownames(coefficients) = predictors
  colnames(coefficients) = c("Estimate","Std.Error","Test statistic","pvalue")
  list(coefficients = coefficients,
       phi = est.full$phi,
       df = est.full$df,
       loglik = loglik.full,
       method = method)
}

NSDE = function(data,predictors,method) {
  x = as.matrix(pData(data))
  y = exprs(data)
  scaling = scalingFactor(data)
  if(is.vector(scaling)) scaling = t(replicate(n,scaling))
  phi = dispersion(data)
  mu.bg = backgroundMean(data)
  phi.bg = backgroundDispersion(data)
  
  if(length(method)==1) method = rep(method,nrow(y))
  if(method[1]=="NanoStringDiff") {
    return(NSDiff(data,predictors))
  }
  out = do.call(rbind,lapply(1:nrow(y), function(i) {
    test = test.func(y[i,,drop=F],x,phi[i],mu.bg,phi.bg,scaling[i,],predictors,method[i])
    with(test,data.frame(predictor=rownames(coefficients),coefficients,method=method,stringsAsFactors=F))
    }))
  outList = list()
  for(predictor in predictors) {
    out.sub = out[out$predictor==predictor,colnames(out)!="predictor",drop=F]
    out.sub[,1:2] = log2(exp(1))*out.sub[,1:2]
    rownames(out.sub) = rownames(y)
    outList[[predictor]] = out.sub
  }
  outList
}

NSDE.cut = function(data,predictors,cutoff=10) {
  y = exprs(data)
  probe.mean = rowMeans(y,na.rm=T)
  mu.bg = backgroundMean(data)
  method = ifelse(probe.mean<cutoff*mean(mu.bg,na.rm=T),"Wald","lm.nb")
  NSDE(data,predictors,method)
}

## add this to utils
invalid = function (x) {
  if (missing(x) || is.null(x) || length(x) == 0) 
    return(T)
  if (is.list(x)) 
    return(all(sapply(x, invalid)))
  else if (is.vector(x)) 
    return(all(is.na(x)))
  else return(F)
}
