
#source("SNV_util.R") # assumes that it is already sourced

#' This function is a wraper for the main function enabling by batch processing and mergin
eval.snv <- function(snv.dat
                     , call.rule = make.call.rule()
                     , subtract.background = T
                     , post.est.EM = F
                     , post.est.debias = T){
  
  
  # store probe.sets info to re-attach at the end
  #-----------------------------------------------
  pannot <- attr(snv.dat,"pannot")

  batched.snv.dat <- snv.dat %>% 
    group_by(userdf.batch.id) %>%  # Group by user defined batch
    nest(.key = userdf.batch.id) %>%
    mutate(userdf.batch.id = map(userdf.batch.id,~ .x %>% 
                                   filter(.,is.smp.ref == "N") %>% # Within each user defined batch get the non-refs
                                   group_by (.,analysis.batch) %>% # Break the non-refs by analysis batch (e.g. cartrige)
                                   do(batches = bind_rows(         # rowbind by batch to refs ignoring refs analysis batches  
                                     .x %>% filter(.,is.smp.ref == "Y") %>% select(-analysis.batch), # remove analysis.batch id as it is inaccurate post rbinding
                                     (.) %>%  select(-analysis.batch)))
    )) %>% unnest(userdf.batch.id)
  
  
  # Run analysis by batch
  #----------------------
  batched.snv.dat$batches %>% purrr::map(.,~ eval.batch.snv(snv.dat = .x %>% data.frame
                                                            , call.rule = call.rule
                                                            , subtract.background = subtract.background
                                                            , post.est.EM = post.est.EM
                                                            , post.est.debias = post.est.debias)) -> batched.res
  
  # Merge the output to the native snv out format
  # compatible with visualizaiton functions etc.
  #----------------------------------------------
  names(batched.res) <- batched.snv.dat$analysis.batch
  snv.f2 <- abind:::abind(sapply(X = batched.res,FUN = function(x) x$snv.f2,simplify = F,USE.NAMES = T),along = 1)
  snv.dat <- bind_rows(sapply(X = batched.res,FUN = function(x) x$snv.dat,simplify = F,USE.NAMES = T))
  graph.EM <- lapply(batched.res,FUN = function(x) x$graph.EM)
  
  # Re-attach probe.sets info
  #--------------------------
  pannot -> attr(snv.dat,"pannot")
  
  return(list(snv.f2 = snv.f2,
              snv.dat = snv.dat,
              call.rule = call.rule,
              graph.EM = graph.EM))
  
}






#' This function is the main function estimating and calling snv. It receives and nxp (samplesxprobes) raw data matrix, along with the sample annotations and probe annotations
#' peforms SNV analysis and returns a list containing the estimates, calls and some analysis parameters
#' @param  snv.dat, data.frame validated and formated for the analysis by preprocess.snv
#' @param call.rule: This is a named list used to generate calls based on fold change and pvalue. the parameters needed are: fc.exclude which specifies a range of fold changes 
#' to exclude from calling, sig.level which specifies pvalue threshold all calls need to be below. rule.in.text and rule.out.text each are a label character vector for printing on the plots only.
#' This is made by make.call.rule function
#' @param subtract.background logical if TRUE, estimated background estimated in preprocessing is subtracted before inference
#' @param post.est.EM logical if TRUE, a post estimation step iteratively seeks to improve on the initial calls made, see get.snv.f2.EM for details
#' @param post.est.debias logical if TRUE, a post estimation step seeks to remove possible bias introduced by normalization 

eval.batch.snv <- function(snv.dat
                           , call.rule = make.call.rule()
                           , subtract.background = T
                           , post.est.EM = F
                           , post.est.debias = T){
  
  
  # Rename the generic as grouping
  #-------------------------------
  snv.dat<- snv.dat %>% 
    dplyr:::rename(.,smp.grouping=smp.generic.name) %>% 
    mutate(.,in.ref = c("N","Y")[1+(smp.grouping == "ref")], smp.grouping.EM = smp.grouping)
  
  
  
  # Consruct the mapping sample label mapping object
  # Note: one-to-one mapping has between grouping and 
  #       sample label is already enforced by propocess
  #-----------------------------------------------------
  smp.name.2.lab <- snv.dat %>% 
    select(smp.label,smp.grouping) %>% 
    distinct(smp.grouping) %>% 
    filter(smp.grouping!="ref")
  rownames(smp.name.2.lab) <- smp.name.2.lab$smp.grouping
  
  
  # Set the initial model formula
  #------------------------------
  frm <- as.formula(nlCount ~ smp.grouping)
  if(subtract.background)
    frm <- as.formula( (nlCount - bg.hat) ~ smp.grouping )
  if(all(!is.na(snv.dat$adj.factor)))
    frm <- update.formula(frm, ~ . + adj.factor)
  
  # if TRUE instead of testing whether the means are different, the test will be the likelihood of data under
  # reference density distribution. Basic model currently uses the mean difference test but post-estimations
  # use prediction sd and thereby effectively doing the likelihood test. Right now if EM is run, the subsequent
  # steps will be based on likelihood test.
  use.likelihood.test <- FALSE 
  
  # Initialize the storage for fold change stats
  #---------------------------------------------
  fc.stats <- snv.dat %>% 
    filter(., smp.grouping!="ref") %>% 
    select(.,c(smp.grouping,probe.label,count)) %>%
    mutate(.,estimate=NA, se=NA, resid.scale = NA,pred.sd = NA,t.value=NA, p=NA, ci.lo=NA, ci.hi=NA, resid.dof=NA, snv.call=NA,tmp.id = paste(smp.grouping,probe.label,sep=".")) %>% 
    arrange(.,probe.label,smp.grouping)
  

    
  
  for(probe.i in unique(snv.dat$probe.label)){
    dat.pi <- filter(snv.dat,probe.label == probe.i)
    dat.pi$smp.grouping <- relevel(as.factor(dat.pi$smp.grouping),ref = "ref")
    lmod.pi <- lm(frm, data = dat.pi)
    sm.pi <- summary(lmod.pi)
    fc.stats.pi <- as.data.frame(sm.pi$coefficients)
    fc.stats.pi <- cbind(fc.stats.pi,
                         matrix(confint(lmod.pi,level=1-call.rule$sig.level),ncol = 2,dimnames = list(rownames(fc.stats.pi),c("ci.lo", "ci.hi")))) # note this uses t stat quantiles
    fc.stats.pi <- fc.stats.pi[!rownames(fc.stats.pi) %in% c("(Intercept)","adj.factor"),]
    fc.stats.pi <- cbind(fc.stats.pi,resid.dof=lmod.pi$df.residual, resid.scale = sd(lmod.pi$residuals))
    fc.stats.pi$pred.sd <- sqrt(fc.stats.pi$`Std. Error`^2 + fc.stats.pi$resid.scale^2)
    fc.stats.pi$tmp.id <- paste(gsub("smp.grouping","",rownames(fc.stats.pi)),probe.i,sep=".")
    fc.stats[match(fc.stats.pi$tmp.id,fc.stats$tmp.id),c("estimate", "se","resid.scale","pred.sd", "t.value", "p","ci.lo", "ci.hi", "resid.dof")] <-
      fc.stats.pi[,c("Estimate", "Std. Error","resid.scale","pred.sd", "t value", "Pr(>|t|)","ci.lo", "ci.hi", "resid.dof")]
  }
  
  rm(fc.stats.pi)
  fc.stats$tmp.id <- NULL
  
  
  # Convert p-value to one sided test
  #-----------------------------------
  fc.stats <- fc.stats %>% 
    mutate(., p = pt(q = .$t.value, df = .$resid.dof,lower.tail = F))
  
  # Construct the snv.f2 array
  #---------------------------
  fc.stats %>% 
    gather(.,key = stat.type,value= value,-c(smp.grouping, probe.label)) %>% 
    reshape2:::acast(., smp.grouping ~ probe.label ~ stat.type,drop = F) -> snv.f2
  
  snv.f2 <- snv.f2[,,c("count","estimate", "se","resid.scale","pred.sd", "t.value", "p","ci.lo", "ci.hi", "resid.dof","snv.call"), drop=F]
  
  
  # Make calls
  #-----------
  snv.calls <- make.snv.call(snv.f2 = snv.f2,call.rule = call.rule)
  snv.f2[,,"snv.call"] <- snv.calls$snv.call[dimnames(snv.f2[,,"snv.call",drop = F])[[1]], dimnames(snv.f2[,,"snv.call", drop = F])[[2]], drop = F ]
  
  data.frame(smp.grouping = rownames(snv.calls$snv.call),snv.calls$snv.call,check.names = F,stringsAsFactors = F) %>%
    gather(.,key = probe.label,value = snv.call0,-smp.grouping) %>% left_join(x=snv.dat,y=.,by=c("smp.grouping","probe.label")) %>%
    mutate(.,snv.call = snv.call0) -> snv.dat
  
  
  
  
  # Post Est. EM
  #-------------
  if(post.est.EM){
    use.likelihood.test <- TRUE
    EM.results <- get.snv.f2.EM(frm,snv.dat,call.rule = call.rule,visualize = F)
    snv.f2 <- EM.results$snv.f2
    snv.dat <- EM.results$snv.dat
    graph.EM <- EM.results$g
  }
  
  
  
  # Do post estimate debiasing
  #---------------------------
  if(post.est.debias){
    snv.f2 <- debias.by.sample(snv.f2 = snv.f2,use.likelihood.test = use.likelihood.test)$snv.f2
    
    #Re-do calling
    snv.calls <- make.snv.call(snv.f2 = snv.f2,call.rule = call.rule) 
    snv.f2[,,"snv.call"] <- snv.calls$snv.call[dimnames(snv.f2[,,"snv.call", drop = F])[[1]], dimnames(snv.f2[,,"snv.call", drop = F])[[2]], drop = F ]
    
  }
  
  # Add calls to dfl.informative
  #-----------------------------
  data.frame(smp.grouping = rownames(snv.calls$snv.call),snv.calls$snv.call,check.names = F,stringsAsFactors = F) %>%
    gather(.,key = probe.label,value = snv.call.debias,-smp.grouping) %>% left_join(x=snv.dat,y=.,by=c("smp.grouping","probe.label")) %>%
    mutate(.,snv.call = snv.call.debias) -> snv.dat
  
  out <- list(snv.f2 = snv.f2, snv.dat = snv.dat, call.rule = call.rule)
  if(post.est.EM)
    out$graph.EM <- graph.EM
  
  return(out)
}