#' This funciton is a wrapper for plot.summary.snv and plot.by.sample.snv
#' @param snv  The list output of snv.f2 which contains the pxnx5 array of estimates
#' @param sample.name A character vector of sample names to be ploted. When NULL, all samples will be ploted
#' @param path The path where the plots are to be outputted. When NULL, the plots are outputted to console
#' @param sample.rename An ordered character vector of sample renames used to label plots. It needs to match the length of sample.names. When NULL, it'll be the same as sample.names
#' @param file.extension character image type to store data as (e.g. png). Note csv generates the csv version of the plotted content
#' @return  NULL

plot.snv <- function(snv, sample.name = NULL ,path = NULL, sample.rename = NULL, file.extension = "pdf", type = c("summary","by.sample","QC")){
  
  type <- match.arg(type,c("summary","by.sample","QC"),several.ok = T)
  
  # Validate input
  #---------------
  tmp <- validate.snv(snv = snv, sample.name = sample.name, sample.rename = sample.rename)
  snv <- tmp$snv
  sample.name <- tmp$sample.name
  sample.rename <- tmp$sample.rename
  
  snv.f2 <- snv$snv.f2
  
  
  prb.names <- dimnames(snv.f2)[[2]]
  
  fls <- list()
  if("summary" %in% type)
    fls$summary <- plot.summary.snv(snv = snv,
                                    sample.name = sample.name ,
                                    path = path,
                                    sample.rename = sample.rename,
                                    file.extension = file.extension)
  
  if("by.sample" %in% type)
    fls$samples <- plot.by.sample.snv(snv = snv,
                                      sample.name = sample.name ,
                                      path = path,
                                      sample.rename = sample.rename,
                                      file.extension = file.extension)
  
  if("QC" %in% type)
    fls$QC <- plot.QC.snv(snv = snv,
                          sample.name = sample.name ,
                          path = path,
                          sample.rename = sample.rename,
                          file.extension = file.extension)
  
  return(fls)
  
}



plot.summary.snv <- function(snv, sample.name = NULL, path = NULL, sample.rename = NULL, file.extension = "pdf"){
  
  
  # Validate snv input and append color to call.rule
  #-------------------------------------------------
  snv$call.rule$categories <- append.color.to.call.rule(call.rule = snv$call.rule)$categories
  
  
  # Convert snv.f2 from array to dat.frame
  #---------------------------------------
  call.df <- get.snv.call.summary.mat(snv = snv, sample.name = sample.name, sample.rename = sample.rename)
  
  call.df %>%
    ggplot(., aes(x=probe.label, y= smp.label,  fill = snv.call)) + 
    geom_tile(colour = "white",size = 1)+
    ggtitle("SNV call summary")+
    ylab("Samples") + 
    xlab(NULL) +
    scale_fill_manual(values = alpha(colour =  snv$call.rule$categories$color, alpha =   snv$call.rule$categories$alpha ), labels =  snv$call.rule$categories$label, drop = FALSE)+
    theme(panel.grid.major = element_blank(), 
          panel.grid.minor = element_blank(),
          axis.text.x = element_text(angle = 90, hjust = 1), 
          plot.title = element_text(size = 20, face = "bold"),
          legend.key.size = unit(.5, "cm"),
          legend.text = element_text(size = 15),
          legend.title = element_text(size = 15),
          legend.position = "bottom") +
    guides(fill=guide_legend(title = NULL,nrow = nrow( snv$call.rule$categories),byrow = TRUE,  title.position = "top"))-> plt
  
  
  if(!is.null(path))
    dir.create(path = path,showWarnings = F)
  
  
  # Convert from long to  wide matrix
  #----------------------------------
  call.df<- call.df %>%
    spread(.,key = probe.label, value = snv.call) %>%
    `rownames<-` (.$smp.label) %>%
    select(-smp.label)
  
    
  # Store direcotry
  #----------------
  fls <- c()
  
  # pdf
  if("pdf" %in% file.extension){
    fl.i <- paste0(path,"/SNV_summary.pdf")
    pdf.path <- pdf(file = fl.i, width = 25, height = max(15*nrow(call.df)/48,12),pointsize = 10)
    print(plt)
    dev.off()
    fls <- c(fls,fl.i)
  }
  
  # png
  if("png" %in% file.extension){
    fl.i <- paste0(path, "/SNV_summary.png")
    png(filename = fl.i,width = 5000,height = max(3500*nrow(call.df)/48,2500),res = 250)
    print(plt)
    dev.off()
    fls <- c(fls,fl.i)
  }
  
  # csv
  if("csv" %in% file.extension){
      fl.i <-  paste0(path, "/SNV_summary.csv")
      # data.frame(snv$snv.f2[,,"snv.call"],check.names = F) %>% `rownames<-` (sample.rename[rownames(.)]) %>%
      write.csv(call.df,file = fl.i)
      fls <- c(fls,fl.i)

  }
    
  class(fls) <- c(class(fls),"snv.files")
  return(fls)
  
}

plot.by.sample.snv <- function(snv, sample.name = NULL ,path = NULL, sample.rename = NULL, file.extension = "pdf"){
  
  # Validate input
  #---------------
  tmp <- validate.snv(snv = snv, sample.name = sample.name, sample.rename = sample.rename)
  snv <- tmp$snv
  sample.name <- tmp$sample.name
  sample.rename <- tmp$sample.rename
  
  snv.f2 <- snv$snv.f2


  prb.names <- dimnames(snv.f2)[[2]]

  
  # exclude probes not estimated (usually invariant contorl)
  #----------------------------------------------------------
  prb.names <- prb.names[prb.names %in% dimnames(snv.f2)[[2]]]
  
  
  # map and append call.type to color
  #---------------------------------
  call.rule <- append.color.to.call.rule(call.rule = snv$call.rule)$categories
  
  
  # For each sample in sample.name plot
  #------------------------------------
  fld.change.limits.min <- c(-4,6)
  
  sapply(sample.name, FUN = function(x) NA,simplify = F,USE.NAMES = T)-> plts
  for(smp.i in sample.name){
    
    snp.smp.i <- data.frame(snv.f2[smp.i, prb.names,])
    
    snp.smp.i %>% mutate(color = call.rule$color[1+(.$snv.call)],
                         snv.call = factor(call.rule$call.type[1+(.$snv.call)],levels = call.rule$call.type)) -> snp.smp.i
    snp.smp.i$fc.label <- ""
    snp.smp.i$fc.label[snp.smp.i$snv.call != call.rule$call.type[1]] <- sprintf("%1.1f",snp.smp.i$estimate[snp.smp.i$snv.call != call.rule$call.type[1]])
    snp.smp.i$xlabels <- prb.names
    snp.smp.i %>% 
      ggplot(data= .,aes(x= xlabels,y=estimate,fill = snv.call, label = fc.label)) +
      geom_crossbar(aes(ymin = ci.lo,ymax = ci.hi),colour = "white",size=.2)+
      geom_text(angle = 90)+
      theme_bw()+ 
      theme(axis.text.x=element_text(angle=90, hjust=1))+
      scale_fill_manual(values = alpha(colour = call.rule$color, alpha =  call.rule$alpha ), labels = call.rule$label, drop=FALSE)+
      geom_hline(yintercept = call.rule$fc.lo[-1],linetype = "dashed",colour = call.rule$color[-1],size = 1)+
      geom_hline(yintercept = 0) +
      ggtitle( sample.rename[smp.i])+
      xlab (NULL)+
      ylab ("log2 Ratio Relative to Ref.")+ 
      theme(axis.text.x=element_text(color = snp.smp.i$color),legend.position = "bottom")+
        guides(fill=guide_legend(title = NULL,nrow = nrow(call.rule),byrow=TRUE,  title.position = "top")) +
      geom_hline( yintercept = fld.change.limits.min, colour = NA) -> plts[[smp.i]] 
    
  }
  
  
  if(!is.null(path))
    dir.create(path = path,showWarnings = F)
  
  # Store direcotry
  #----------------
  fls <- c()

  
  # pdf
  if("pdf" %in% file.extension){
    fl.i <- paste0(path,"/SNV_plots.pdf")
    pdf.path <- pdf(file = fl.i, width = 19)
    plts
    dev.off()
    fls <- c(fls,fl.i)
  }
  
  # png
  if("png" %in% file.extension){
    for(smp.i in sample.name){
     
      fl.i <- paste0(path, "/",swap.illegal.char (sample.rename[smp.i]),".png")
      png(filename = fl.i,width = 3000,height = 1800,res = 200)
      print(plts[[smp.i]])
      dev.off()
      fls <- c(fls,fl.i)
    }
  }
  
  if("csv" %in% file.extension){
    for(smp.i in sample.name){

      fl.i <- paste0(path, "/",swap.illegal.char (sample.rename[smp.i]),".csv")
      snv.f2[smp.i,,c("count","estimate","se","t.value","p","ci.lo","ci.hi","snv.call")] %>% 
        write.csv(.,file = fl.i)
      fls <- c(fls,fl.i)
    }
  }



  
  # Other plots to be added
  
  class(fls) <- c(class(fls),"snv.files")
  return(fls)
  
}

plot.QC.snv <- function(snv,sample.name = NULL ,path = NULL, sample.rename = NULL, file.extension = "pdf"){
  
  attr(snv$snv.dat,"pannot") %>% select(c(probeID, probe.label,control.type)) %>% dplyr::rename(probe = probeID) -> pannot
  ctrl.prbs <- colnames(snv$snv.dat)[colnames(snv$snv.dat) %in% pannot$probe]
  snv.qc.dat <- snv$snv.dat[!duplicated(snv$snv.dat$unique.lane.id),colnames(snv$snv.dat) %in% c("unique.lane.id","cartridge.id", "analysis.batch", "smp.label", "smp.grouping", "is.smp.ref", "adj.factor",ctrl.prbs)]
  ctrl.labs <- pannot$probe.label[match(colnames(snv.qc.dat)[colnames(snv.qc.dat) %in% ctrl.prbs],pannot$probe)]
  ctrl.types <- data.frame(type = pannot$control.type[match(colnames(snv.qc.dat)[colnames(snv.qc.dat) %in% ctrl.prbs],pannot$probe)],row.names = ctrl.labs)
  colnames(snv.qc.dat)[colnames(snv.qc.dat) %in% ctrl.prbs] <- ctrl.labs
  rownames(snv.qc.dat) <- snv.qc.dat$unique.lane.id
  
  
  if(!is.null(path))
    dir.create(path = path,showWarnings = F)
  
  
  # Store direcotry
  #----------------
  fls <- c()
  
  # pdf
  if("pdf" %in% file.extension){
    fl.i <- paste0(path,"/SNV_QC_raw.pdf")
    pdf.path <- pdf(file = fl.i, width = max(15,15*nrow(snv.qc.dat)/24), height = max(15*length(ctrl.labs)/17,15),pointsize = 30,onefile=FALSE)
    pheatmap(mat = t(log10(snv.qc.dat[,ctrl.labs])),cluster_rows = F,cluster_cols = F,scale = "none",main = "log10 Raw SNV Control Probes",display_numbers = T,annotation_col =  snv.qc.dat[,"is.smp.ref",drop=F],annotation_row = ctrl.types)
    # plot(1,1)
    dev.off()
    fls <- c(fls,fl.i)
    
    fl.i <- paste0(path,"/SNV_QC_scaled.pdf")
    pdf.path <- pdf(file = fl.i, width = max(15,15*nrow(snv.qc.dat)/24), height = max(15*length(ctrl.labs)/17,15),pointsize = 30,onefile=FALSE)
    pheatmap(mat = t(log10(snv.qc.dat[,ctrl.labs])),cluster_rows = F,cluster_cols = F,scale = "none",main = "log10 Raw SNV Control Probes",display_numbers = T,annotation_col =  snv.qc.dat[,"is.smp.ref",drop=F],annotation_row = ctrl.types)
    dev.off()
    fls <- c(fls,fl.i)
  }
  
  # png
  if("png" %in% file.extension){
    fl.i <- paste0(path, "/SNV_QC_raw.png")
    png(filename = fl.i,width = max(3500,3500*nrow(snv.qc.dat)/24) ,height = max(3500*length(ctrl.labs)/17,2000),res = 300)
    pheatmap(mat = t(log10(snv.qc.dat[,ctrl.labs])),cluster_rows = F,cluster_cols = F,scale = "none",main = "log10 Raw SNV Control Probes",display_numbers = T,annotation_col =  snv.qc.dat[,"is.smp.ref",drop=F],annotation_row = ctrl.types)
    dev.off()
    fls <- c(fls,fl.i)
    
    fl.i <- paste0(path, "/SNV_QC_scaled.png")
    png(filename = fl.i,width = max(3500,3500*nrow(snv.qc.dat)/24) ,height = max(3500*length(ctrl.labs)/17,2000),res = 300)
    pheatmap(mat =t(  scale(log10(snv.qc.dat[,ctrl.labs]),center =  colMedians(log10(as.matrix(snv.qc.dat[,ctrl.labs]))),scale = F)),cluster_rows = F,cluster_cols = F,scale = "none",main = "log10 Median Centered SNV Control Probes",display_numbers = T,annotation_col =  snv.qc.dat[,"is.smp.ref",drop=F],annotation_row = ctrl.types)
    dev.off()
    fls <- c(fls,fl.i)
  }
  
  # csv
  if("csv" %in% file.extension){
    fl.i <- paste0(path, "/SNV_QC_raw.csv")
    as.data.frame(t(log10(snv.qc.dat[,ctrl.labs])), check.names =  F) %>% 
      write.csv(.,file = fl.i)
    fls <- c(fls,fl.i)
    
    fl.i <- paste0(path, "/SNV_QC_scaled.csv")
    as.data.frame(t(  scale(log10(snv.qc.dat[,ctrl.labs]),center =  colMedians(log10(as.matrix(snv.qc.dat[,ctrl.labs]))),scale = F)), check.names =  F) %>% 
      write.csv(.,file = fl.i)
    fls <- c(fls,fl.i)
    
  }
  
  
  class(fls) <- c(class(fls),"snv.files")
  return(fls)
  
}




#'This function maps call categories defined in call.rule data.frame to color-blind friendly colors
#'@param call.rule this is the list generated by make.call.rule
#'@return call.rule with the color appended.

append.color.to.call.rule <- function(call.rule){
  
  colpalette <- cbpalette <- c("#999999", "#56B4E9","#E69F00", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7")
  
  n <- n_distinct(call.rule$categories$call.type)
  if(n!=nrow(call.rule$categories))
    stop("The SNV call.rule$categories are supposed to be unique (no duplicate is allowed)")
  
  # If the number of categories exceeds number in the colorbind palette revert to default coloring scheme
  if(n>length(cbpalette))
    colpalette <- hcl(h = seq(15, 375, length = n + 1),l = 65, c = 100)
  
  call.rule$categories$color <- colpalette[1:nrow(call.rule$categories)]
  call.rule$categories$alpha <- 1
  call.rule$categories$alpha[1] <- 0.5
  
  return(call.rule)
}

#' #' This funciton of a plot only utility funciton
#' #' @param string a string in which there are illegal characters
#' #' @param illegal.pattern Regx pattern of characters separated by "|" to be replaced
#' #' @param replacement a character string to replace the illegal chars
#' 
#' remove.illegal.chars <- function(string,
#'                                  illegal.pattern = paste(c("<",">",":","\"","\\?","\\|","\\*","\\."),collapse = "|"),
#'                                  replacement = "_"){
#'   smp.i.rename <- gsub(x = string,pattern = illegal.pattern,replacement = replacement)
#' }

#' This function receives a character vector and removes 
#' Illegal characters that break the analysis (either in html or plot saving)
#' NOTE: Decided against adding the 255 char limit check as that was very unlikely for geneset name
#' @param x a character vator

swap.illegal.char <- function(x){
  char.2b.changed <- c(STR = "\\*",
                       FR  = "<",
                       TO  = ">",
                       CL  = "\\:",
                       TLD = "\\~",
                       BNG = "\\!",
                       AT  = "\\@",
                       DLR  = "\\$",
                       PRC = "\\%",
                       CR  = "\\^",
                       AP  = "\\`",
                       EQ  = "\\=",
                       CBL = "\\{",
                       CBR = "\\}",
                       BRL = "\\[",
                       BRR = "\\]",
                       PP = "\\|",
                       QZ = "\\?",
                       PLS = "\\+",
                       SQT = "\\'",
                       DQT = "\\\"",
                       FS = "\\/",
                       BS = "\\\\")
  
  # Remove .png and .RCC
  x <- gsub("\\.png|\\/|\\.RCC","",x)
  
  # Test case
  # x <- "asdf * < > : ~ ! @ $ % ^ ` = { } [ ] | ? + ' \" / \\ "
  # x
  for(i in 1:length(char.2b.changed)){
    pt <- unname(char.2b.changed[i])
    rpl <- paste(".",names(char.2b.changed)[i],".",sep="")
    x <- gsub(pt,rpl,x = x)
    # cat(pt,"\n")
    # cat(x,"\n")
  }
  return(x)
}