#' This function generates generic sample id names to standaridze the internal data wrangling
#'  @param sannot nxs data.frame of sample annotations containing the identity of samples to be used as reference 
#'  @param replicate.id.column if there are samples run at replicated, the column name (character) identifying those samples. NULL will run each sample as singlton
#'  @param ref.indices the row indices corresponding to the reference samples

get.generic.smp.names <- function(sannot
                                  , replicate.id.column = NULL
                                  , ref.indices){
  if(!is.numeric(ref.indices))
    stop("ref.indices need to be numeric")
  
  smp.rep.id <- rownames(sannot)
  if(!is.null(replicate.id.column))
    smp.rep.id <- sannot[,replicate.id.column]
  
  # generify names
  #---------------
  smp.generic.rep.id <-   smp.rep.id
  smp.generic.rep.id[ref.indices] <- "ref"
  smp.generic.rep.id <- factor(smp.generic.rep.id,levels = c("ref",unique(smp.generic.rep.id[smp.generic.rep.id!="ref"])))
  smp.generic.rep.id <- sprintf("sample%04d", as.numeric(smp.generic.rep.id)-1)
  smp.generic.rep.id[smp.generic.rep.id==sprintf("sample%04d",0)] <- "ref"
  
  
  return(smp.generic.rep.id)
}




#' This function receives and validates the raw SNV count matrix and the associated sample annotation and
#' probe annotation as well as the estimation options and parameters. It then performs basic
#' pre-processing of the data making it ready for inference
#' @param raw.dat nxp data matrix of n samples and p probes raw counts
#' @param sannot: an nxk dataframe of sample annoations (as characters). The rownames match those of the raw.dat.
#' @param pannot: a pxj datafarme of probe annotations (as character). The rownames match the colnames of raw.dat. pannot needs to contain columns Gene.Name (a character vector), 
#' and is.control(a"Y","N" character vector)
#' @param loci.to.estimate: a character vector of probes for which SNV analysis is to be performed. By default all non control probes If NULL
#' @param smp.names: a character vector of (non-reference) sample names for which we would like to perform SNV analysis. The names need to be a subset of those in the rownames of the raw.data 
#' @param smps.ref: a character vector of samples names to be used as reference (i.e. WT). The names need to be a subset of those in the rownames of raw.dat
#' @param adj.frm a character formula allowing computation of the adjustment factors based on control probes. When NULL, no adjustment will be made
#' @param replicate.id.column if there are samples run at replicated, the column name (character) identifying those samples. NULL will run each sample as singlton
#' @param smp.relabels optional mapping samples to new labels
#' 
#' NOTE: rownames of sannot are expected to be unique per pre-processing in load_n_clean
preprocess.snv <- function(raw.dat
                           , sannot
                           , pannot
                           , loci.to.estimate = NULL
                           , smp.names
                           , samps.ref
                           , adj.frm = NULL
                           , replicate.id.column = NULL
                           , smp.relabels = NULL
                           , analysis.batch = c("all","cartridge.id")){
  
  
  analysis.batch  <- match.arg(analysis.batch,  c("all","cartridge.id"),several.ok = F)
  
  
  
  
  #Ensure the inputs are valid
  #---------------------------
  
  pannot.colnames <- c(probe.label = "Probe.Label",
                       probeID= "ProbeID",
                       LOC = "SNV.LocID",
                       probe.type = "SNV.Probe.Type",
                       is.control = "Is.Control",
                       control.type = "Control.Type")
  
  
  
  # Check pannot
  if(any(!pannot.colnames %in% colnames(pannot)))
    stop(paste("Missing the following required probe annotations:",
               paste(pannot.colnames[!pannot.colnames %in% colnames(pannot)],collapse = ",")))
  
  # Rename pannot
  colnames(pannot)[match(pannot.colnames,colnames(pannot))] <- names(pannot.colnames)
  pannot %>% mutate(.,is.SNV.Ref = c("Y","N")[1+ (.$probe.type != "SNV_REF") ] )-> pannot
  pannot[pannot$control.type!="","probe.type"] <- pannot[pannot$control.type!="","control.type"]
  pannot %>% mutate(.,is.control = c("Y","N")[1+ (.$is.control == 0) ] )-> pannot
  

  
  if(!"SNV.norm.probeset" %in% colnames(pannot))
  {

    pannot$"SNV.norm.probeset"  <- NA
    pannot$SNV.norm.probeset[pannot$probe.type %in% c("Positive","SNV_PCR_CTL","SNV_POS","SNV_INPUT_CTL","SNV_REF")] <-
      pannot$probe.type[pannot$probe.type %in% c("Positive","SNV_PCR_CTL","SNV_POS","SNV_INPUT_CTL","SNV_REF")]
  }
  if(!"SNV.norm.probeset.use"%in% colnames(pannot)){
    pannot$"SNV.norm.probeset.use" <- NA
    pannot$SNV.norm.probeset.use[pannot$probe.type %in% c("SNV_PCR_CTL")] <- "SNV_POS"
    pannot$SNV.norm.probeset.use[pannot$probe.type %in% c("SNV_POS","SNV_NEG","SNV_UDG_CTL")] <- "SNV_POS" # this also has the temp but using the ratio of raw it doesn't matter
    pannot$SNV.norm.probeset.use[pannot$probe.type %in% c("SNV_INPUT_CTL")] <- "SNV_PCR_CTL"  #"Positive;SNV_PCR_CTL"
    pannot$SNV.norm.probeset.use[pannot$probe.type %in% c("SNV_REF","SNV_VAR")] <- "SNV_INPUT_CTL"
  }
  
  pannot.colnames <- c(pannot.colnames, 
                       c(is.SNV.Ref ="is.SNV.Ref",
                         SNV.norm.probeset = "SNV.norm.probeset",
                         SNV.norm.probeset.use = "SNV.norm.probeset.use"))
  
  
  if(any(rownames(sannot)!= rownames(raw.dat)))
    stop("Rownames of the provided sample annotations do not match those of raw data")
  
  if(any(!smp.names %in% rownames(raw.dat)))
    stop("Selected samples are not in rownames of the raw data")
  
  if(any(!samps.ref %in% rownames(raw.dat)))
    stop("Selected reference samples are not in rownames of the raw data")
  
  
  # Convert probeID to be R-friendly
  #----------------------------------
  pannot$probeID <- make.names(pannot$probeID)
  colnames(raw.dat) <- make.names(colnames(raw.dat))
  
  #trim data to include only smp.names and samps.ref
  #-------------------------------------------------
  raw.dat <- raw.dat[c(samps.ref,smp.names),,drop=F]
  sannot <- sannot[c(samps.ref,smp.names),,drop=F]
  
  
  # what happens if somebody doesn't choose any of the WT to be in the loci.to.estimate? note wT are added as features below
  # or what happens if somebody picks controls as loci to estimate --> need better vetting of the choice!
  if(is.null(loci.to.estimate)) 
    loci.to.estimate <- pannot$probeID[pannot$is.control=="N"]
  
  smp.grp <- data.frame(row.names = smp.names,smp.grp = smp.names,stringsAsFactors = F)
  if(!is.null(replicate.id.column))
    smp.grp <- data.frame(smp.grp=sannot[smp.names,replicate.id.column])
  
  
  if(is.null(smp.relabels))
    smp.relabels<- data.frame(row.names = smp.names,smp.label = smp.grp[smp.names,"smp.grp"],stringsAsFactors = F)
  
  if(any(apply(table(smp.grp$smp.grp,smp.relabels$smp.label)!=0,1,sum)!=1)){
    warning("SNV sample label and sample replicate id do not have a one-to-one relation. sample grouping is used as labels")
    smp.relabels<- data.frame(row.names = smp.names,smp.label = smp.grp[smp.names,"smp.grp"],stringsAsFactors = F)
  }
  rm(smp.grp)
  
  
  #Define analysis batch per function input
  #----------------------------------------
  sannot$cartridge.id <- "non.ref.cartridges"
  sannot[samps.ref,"cartridge.id"] <- "ref.cartridges"
  sannot$analysis.batch <- sannot$cartridge.id
  if(!is.null(sannot$run.batch.id))
    sannot$cartridge.id <- sannot$run.batch.id
  
  if(analysis.batch == "cartridge.id")
    sannot$analysis.batch <- sannot$cartridge.id
  
  
  #Append to sannot label and generic names (latter contains rep info ==> may be non-unique)
  #----------------------------------------------------------------------------------------
  sannot$unique.lane.id <- rownames(sannot)
  
  sannot[rownames(smp.relabels),"smp.label"] <- smp.relabels$smp.label
  sannot$smp.generic.name <- get.generic.smp.names(sannot = sannot,
                                                   replicate.id.column = replicate.id.column,
                                                   ref.indices = which(rownames(raw.dat) %in% samps.ref))
  sannot$userdf.batch.id.labels <- sannot$Batch.ID
  sannot$Batch.ID <- as.character(factor(sannot$Batch.ID,levels = unique(sannot$Batch.ID), paste("userdf.batch",1:n_distinct(sannot$Batch.ID),sep=".")))
  
  # here we drop all the unnecessary sannots
  sannot %>% select(.,c(unique.lane.id,cartridge.id,Batch.ID,userdf.batch.id.labels,analysis.batch,smp.label,smp.generic.name)) %>%
    dplyr::rename(userdf.batch.id = Batch.ID) %>%
    mutate(.,is.smp.ref = 1+ as.numeric(smp.generic.name == "ref")) %>%
    mutate(.,is.smp.ref = c("N","Y")[is.smp.ref]) -> sannot
  
  
  
  # Append to sannot adjustment factor (temperature)
  #-------------------------------------------------
  sannot$adj.factor <- NA
  if(!is.null(adj.frm)){
    trms <- as.character(attr(terms(adj.frm),"variables"))[-1]
    re.trms <- pannot$probeID[match(trms,pannot$probe.label)]
    names(re.trms) <- trms
    frm <- as.character(adj.frm)[-1]
    for(trm in trms){
      frm <- gsub(pattern = trm,replacement = re.trms[trm],x = frm)
    }
    adj.frm <- as.formula(paste0("~",frm))
    sannot$adj.factor <- log2(f_eval(adj.frm,data.frame(raw.dat+1)))
  }
  
  
  # Append  to sannot normalization factors
  #----------------------------------------
  norm.probe.sets <- split(pannot$probeID,f = pannot$SNV.norm.probeset)
  sannot <- data.frame(sannot,scale(as.data.frame(lapply(norm.probe.sets,function(x)rowMeans(log2(raw.dat[,x]+1)))),center = T,scale = F)) 
  
  
  # Append SNV controls to the sannot
  #----------------------------------
  probe.sets <- split(pannot$probeID,f = pannot$probe.type)
  sannot <- data.frame(sannot,raw.dat[rownames(sannot),c(probe.sets$SNV_INPUT_CTL,
                                                         probe.sets$SNV_UDG_CTL,
                                                         probe.sets$SNV_PCR_CTL,
                                                         probe.sets$SNV_NEG,
                                                         probe.sets$SNV_POS)],stringsAsFactors = F,check.names = F)
  
  # Merge with counts and convert to long
  #--------------------------------------
  dat <- data.frame(unique.lane.id=rownames(raw.dat),raw.dat,check.names = F, stringsAsFactors = F)
  dat <- gather(data= dat,key = probeID, value = count, -unique.lane.id)
  dat <- left_join(x = sannot, y = dat, by = "unique.lane.id")
  
  # Merge with probe annoation
  #---------------------------
  dat <- left_join(x = dat,y = pannot[,names(pannot.colnames)],by="probeID")
  tmp <- dat$count; dat$count <- NULL; dat$count <- tmp; rm(tmp)
  
  # Make WT counts as features
  #---------------------------
  tmp <- split(dat,f = dat$is.SNV.Ref)
  tmp$Y <- tmp$Y[,c("unique.lane.id","LOC","count")]
  colnames(tmp$Y) <- c("unique.lane.id","LOC","WT.count") 
  dat <- left_join(x = tmp$N,y = tmp$Y, by = c("unique.lane.id","LOC"))
  rm(tmp)
  
  # Add normalized counts and WT counts
  #------------------------------------
  tmp <- as.matrix(dat[,unique(dat$SNV.norm.probeset.use)])
  dat$nlCount <-  log2(dat$count + 1) - tmp[cbind(1:nrow(dat),match(dat$SNV.norm.probeset.use,colnames(tmp)))]
  dat$WT.nlCount <-  log2(dat$WT.count + 1) - tmp[cbind(1:nrow(dat),match(dat$SNV.norm.probeset.use,colnames(tmp)))]
  rm(tmp)
  
  # Center WT.nlCount by LOC mean and within each of the is.smp.ref groups
  # This makes us only interested in adjusting for variability of each WT within ref relative to
  # within non-ref as opposed to batch effect which may result in whole-sale bias between the ref/no ref
  #-----------------------------------------------------------------------------------------------------
  dat$WT.nlCount.adj <- NA
  dat[!is.na(dat$WT.nlCount),"WT.nlCount.adj"] <- lm(WT.nlCount ~ LOC:is.smp.ref,data=dat[!is.na(dat$WT.nlCount),])$residuals +
    lm(WT.nlCount ~ LOC,data=dat[!is.na(dat$WT.nlCount),])$fitted.values

  
  #Drop rows not in loci.to.estimate
  #-----------------------------
  dat <- dat[dat$probeID %in% loci.to.estimate,]
  
  
  # Add estimated background
  #-------------------------
  dat$bg.hat <- 0
  if(all(!is.na(sannot$adj.factor))){
    
    rsq.th <- 0.6     # background model should have R2 > rseq.th to be used
    z.coverage <- 1.3 # background model should be based on ref sample data that covers +/- z.coverage non-ref data to be used
    med.count.th <- 50 # Only model background when median counts in ref samples > med.count.th
    
    for(p in unique(dat$probeID)){
      tmp <- dat[dat$probeID==p ,]
      
      med.count <- median(tmp[tmp$is.smp.ref == "Y","count"])
      if(med.count<=med.count.th)
        next
      
      sm.stat <- tmp %>% group_by(is.smp.ref) %>% summarize(sd(nlCount),mean(nlCount),min(nlCount),max(nlCount))
      bg.zmax <- (sm.stat[sm.stat$is.smp.ref=="Y","max(nlCount)"]-sm.stat[sm.stat$is.smp.ref=="N","mean(nlCount)"])/
        sm.stat[sm.stat$is.smp.ref=="N","sd(nlCount)"]
      bg.zmin <- (sm.stat[sm.stat$is.smp.ref=="Y","min(nlCount)"]-sm.stat[sm.stat$is.smp.ref=="N","mean(nlCount)"])/
        sm.stat[sm.stat$is.smp.ref=="N","sd(nlCount)"]
      
      if(!(bg.zmin<= -z.coverage & z.coverage<=bg.zmax))
        next
      
      mod <- lm(nlCount~adj.factor + WT.nlCount.adj, data = tmp[tmp$is.smp.ref == "Y",])
      r2 <- summary(mod)$adj.r.squared
      
      if(r2> rsq.th)
        dat[rownames(tmp) ,"bg.hat"] <- predict(mod,newdata = tmp)
    }
  }
  
  
  
  #call= match.call()
  class(dat) <- c(class(dat),"snv")
  attr(dat,"pannot") <- pannot
  return(dat)
}



# Making call rules and  calls functions
#---------------------------------------

#' This function makes call by p-value cut off. Optionally, it adjust p-value for multiplicity correction
#' @param snv.f2 the array result output containig the stats by probe and by sample needed for making calls
#' @param sig.level level of significance
#' @param adjust if TRUE pvalues are adjusted for multiplicity testing

call.by.pval <- function(snv.f2, call.rule, adjust = F){
  p.mat <- snv.call <- abind:::adrop(snv.f2[,,"p", drop = F],drop = c(F,F,T)) 
  
  # Adjust pvalue sample wise
  if(adjust){
    for(smp in dimnames(snv.f2)[[1]]){
      p.mat[smp,] <- p.adjust(snv.f2[smp,,"p"],method = "BH")
    }
  }
  
  # Ensure monotonicly decreasing
  if(!all(diff(call.rule$categories$sig.level.hi)<=0))
    stop("sig.levels in call rule need to be decreasing or flat")
  
  #since duplicate is allowed labeling needs to be adjusted for cut
  #here we are choosing the highest level of stringency we the p value qulifies for
  labels <- apply(table(1:length(call.rule$categories$sig.level.hi),call.rule$categories$sig.level.hi),2,function(x) max(which(x==1))) [as.character(rev(unique(call.rule$categories$sig.level.hi)))]
  
  snv.call[dimnames(snv.f2)[[1]], dimnames(snv.f2)[[2]]]  <- as.numeric(as.character(cut(p.mat, breaks =  rev(unique(c(call.rule$categories$sig.level.hi,0))), labels = labels) ))
  return(snv.call)
}

#' this function makes call by fold change based on a range outside which fold change relative to reference is to be called snv
#' @param snv.f2 the array result output containig the stats by probe and by sample needed for making calls
#' @param no.call.range a numeric vector of size 2 determining the region outside which fold changes are called snv

call.by.foldchange <- function(snv.f2, call.rule){
  # Ensure monotonicly decreasing
  if(!all(diff(call.rule$categories$fc.lo)>0))
    stop("fc levels in call rule need to be strictly increasing")
  
  est.mat <- snv.call <- abind:::adrop(snv.f2[,,"estimate", drop = F],drop = c(F,F,T)) 
  snv.call[dimnames(snv.f2)[[1]], dimnames(snv.f2)[[2]]] <- cut(est.mat,breaks = c(call.rule$categories$fc.lo,Inf), labels = 1:length(call.rule$categories$fc.lo)) 
  return(snv.call)
}

#' this function makes call by fold change based on a range outside which fold change relative to reference is to be called snv
#' @param snv.f2 the array result output containig the stats by probe and by sample needed for making calls
#' @param no.call.range a numeric vector of size 2 determining the region outside which fold changes are called snv

snv.call.by.raw <- function(snv.f2, call.rule){
  # Ensure monotonicly decreasing
  if(!all(diff(call.rule$categories$raw.background.lo)>0))
    stop("raw.background levels in call rule need to be strictly increasing")
  
  count.mat <- snv.call <- abind:::adrop(snv.f2[,,"count", drop = F],drop = c(F,F,T))
  snv.call[dimnames(snv.f2)[[1]], dimnames(snv.f2)[[2]]] <- cut(count.mat,breaks = c(call.rule$categories$raw.background.lo,Inf), labels = 1:length(call.rule$categories$raw.background.lo)) 
  return(snv.call)
}



#' This fucntion wraps the call rule into a standard format used in call and for labeling plots
#' @param fc.breaks a numeric vector of size n breaking the fold change scale into n+1 region.  
#' Note: the first region is forced to included -Inf. If -Inf missing, it'll be appended. Also values will be sorted in increasing order and duplicates are not allowed.
#' @param sig.level level of significance. A vector of same length as fc.breaks.
#' Note: the first value corresponding to the first region must be 1 or 1 will be appended. Also values will be sorted in decreasing order and duplicates are not allowed.
#' @param call.type the label assigned to each region. A character vector of the same length and fc.breaks.
#' Note: the first value corresponding to the first region is forced to be "No". 
#' @param raw.background  a number which is used to determine what's the raw count snv call needs to be above. The raw count threshold is raw.background*2^fc
#' @return call.rule a data.frame containing attributes defining the regions used in making calls and the corresponding labels.

make.call.rule <- function(fc.breaks = c(-Inf,1,2), call.type = c("No","Undetected","Yes"), sig.level = c(1,.05,.05), raw.background = 50){
  
  # Validate fc.breaks
  max.calltype.cat <- 4 # This is the maximum number of call type categoreis allowed to be defined. It is set at 4 as larger numbers seemed not very useful in the SNV analysis context 
  if(!is.numeric(fc.breaks) | sum(!fc.breaks %in% c(-Inf,Inf)) == 0 | any(duplicated(fc.breaks)) | length(fc.breaks) > max.calltype.cat) {
    warning("fc.breaks provided is not a numeric vector of positive length between 1 and 4 unique values. call.rules are set to default\n")
    fc.breaks <- c(-Inf,1,2)
    call.type <- c("No","Undetected","Detected")
    sig.level <- c(1,.05,.05)
    raw.background <- 50
  }
  
  
  # validate sig.level
  if(length(fc.breaks)!= length(sig.level) | !is.numeric(sig.level) | any(sig.level<0 | 1<sig.level) ){
    warning("sig.level needs numerice between 0 and 1 and be of the same length as fc.breaks. call.rules are set to default\n")
    fc.breaks <- c(-Inf,1,2)
    call.type <- c("No","Undetected","Detected")
    sig.level <- c(1,.05,.05)
  }
  
  # validate call.type
  call.type <- as.character(call.type)
  if(length(fc.breaks)!= length(call.type)){
    warning("fc.breaks and call.type need to be of the same length. The call.type will be replaced\n")
    call.type <- paste("fc larger than",fc.breaks)
  }
  
  
  # validate raw.background
  raw.background <- raw.background[1]
  if(!is.numeric(raw.background) | raw.background<0){
    warning("SNV raw.background provided is not a positive numeric value of length 1. It will be replaced with default\n")
    raw.background <- 50
  }
  
  
  # Append lower bound if needed and update
  if(!-Inf %in% fc.breaks){
    fc.breaks <- c(-Inf, fc.breaks)
    call.type <- make.unique(c("No", call.type))
    sig.level <- c(1,sig.level)
  }
  
  
  brk.order <- order(fc.breaks,decreasing = F)
  fc.breaks <- fc.breaks[brk.order]
  call.type <- call.type[brk.order]
  sig.level <- sig.level[brk.order]
  
  if(any(diff(sig.level)>0)){
    warning("sig.level when sorted by fc.break is not monotonically decreasing. It will be replaced\n")
    sig.level <- c(1,rep(0.05,length(fc.breaks[-1])))
  }
  

  
  categories <- data.frame(fc.lo = fc.breaks,
                           fc.hi = c(fc.breaks[-1],Inf),
                           sig.level.lo = c(sig.level[-1],-Inf),
                           sig.level.hi = sig.level,
                           logic = "or",stringsAsFactors = F) %>%
    mutate(logic = replace(logic, fc.hi == Inf,"and"),
           raw.background.lo = raw.background*2^fc.lo,
           raw.background.hi = raw.background*2^fc.hi,
           call.type = call.type) %>% 
    mutate(fc.label = gsub("-Inf <|< Inf","",paste(round(fc.lo,2),"< log2(fc) <", round(fc.hi,2)  ))) %>%
    mutate(raw.count.label = paste(logic,gsub(pattern = "< Inf|^0 <","", paste(round(raw.background.lo),"< raw count <", round(raw.background.hi))))) %>%
    mutate(sig.label = gsub("-Inf <|< 1$","", paste(logic,round(sig.level.lo,2),"< p value <", round(sig.level.hi,2))))
  
  categories$sig.label[categories$sig.level.hi==categories$sig.level.lo] <- paste("and p value <", round(categories$sig.level.hi[categories$sig.level.hi==categories$sig.level.lo],2)) 
  
  categories <- categories %>% mutate(label = paste(call.type,":",fc.label,raw.count.label,sig.label))
  
  return(list(categories = categories, sig.level = min(sig.level),raw.background = raw.background))
}





#' Given the snv.f2 stats and call.rule object, it makes TRUE/FALSE calls for SNV or not
#' @param snv.f2 the array result output containig the stats by probe and by sample needed for making calls
#' @param call.rule: This is a named list used to generate calls based on fold change and pvalue. the parameters needed are: fc.exclude which specifies a range of fold changes 
#' to exclude from calling, sig.level which specifies pvalue threshold all calls need to be below. rule.in.text and rule.out.text each are a label character vector for printing on the plots only.
#' This is made by make.call.rule function

make.snv.call <- function(snv.f2, call.rule = make.call.rule()){
  
  probes <- dimnames(snv.f2)[[2]]
  sample.name <- dimnames(snv.f2)[[1]]
  
  snv.call <- snv.call.fc <-  snv.call.pv <- snv.call.raw <- matrix(-Inf,nrow = length(sample.name), ncol = length(probes),dimnames = list(sample.name,probes))
  
  # Get info to make snv call
  #--------------------------
  if(!is.null(call.rule$categories$fc.lo))
    snv.call.fc <- call.by.foldchange(snv.f2 = snv.f2, call.rule = call.rule)[sample.name,probes, drop = F]
  
  if(!is.null(call.rule$categories$sig.level.lo))
    snv.call.pv <- call.by.pval(snv.f2 = snv.f2,call.rule = call.rule, adjust = F)[sample.name,probes, drop = F]
  
  if(!is.null(call.rule$categories$raw.background.lo))
    snv.call.raw <- snv.call.by.raw(snv.f2 = snv.f2, call.rule = call.rule)[sample.name,probes, drop = F]
  
  snv.call <- pmin(snv.call.raw,pmin(snv.call.fc,snv.call.pv))-1
  
  # #To convert to categorical
  # snv.call[sample.name,probes] <- call.rule$categories$call.type[as.numeric(snv.call)]
  
  return(list(snv.call = snv.call, call.rule = call.rule))
}


# Post Estimation functions
#---------------------------------------

#' This function uses the initial calls embeded withing snv.dat and using the user-defined options, re-estimates the fold changes using as reference
#' the "no call" cases in non-reference samples. This function iteratively makes calls and re-group samples until converages. The aim of this function
#' is to increase statistical detection power for snvs when the number of reference samples are limited.
#' @param frm this is the same as the adj.frm used to compute the adjustemnt score
#' @param  snv.dat this data.frame is the properly formated data input
#' @param call.rule: This is a named list used to generate calls based on fold change and pvalue. the parameters needed are: fc.exclude which specifies a range of fold changes 
#' to exclude from calling, sig.level which specifies pvalue threshold all calls need to be below. rule.in.text and rule.out.text each are a label character vector for printing on the plots only.
#' This is made by make.call.rule function
#' @param visualize this logical if TRUE, returns the diagnostic plots for excuted function for debugging/evaluation purposes

get.snv.f2.EM <- function(frm,
                          snv.dat,
                          call.rule,
                          visualize = F){
  
  
  # In EM values larger than max.bg.count won't be
  # brought in and added to the reference population
  # this may in the future be adaptively learned from
  # data by probe rather than a fix value.
  max.bg.count <- 1000
  
  # update formula (replace grouping by EM gropuing)
  #-------------------------------------------------
  # frm <- frm0 <- update.formula(frm,.~ .- smp.grouping + smp.grouping.EM + WT.nlCount.adj)
  frm <- frm0 <- update.formula(frm,.~ .- smp.grouping + smp.grouping.EM)
  
  # update the in.ref and initialize the storage for fold change stats
  #-------------------------------------------------------------------
  snv.dat <- snv.dat %>% mutate(., in.ref.tmp =  c("N","Y")[1+(in.ref == "Y"| snv.call != max(snv.call,na.rm = T))])
  snv.dat[snv.dat$in.ref.tmp =="Y","smp.grouping.EM"] <- "ref"
  
  fc.stats <- snv.dat %>% 
    filter(., in.ref=="N") %>% 
    select(.,c(smp.grouping,probe.label,count)) %>%
    mutate(.,estimate=NA, se=NA, resid.scale = NA,pred.sd = NA,t.value=NA, p=NA, ci.lo=NA, ci.hi=NA, resid.dof=NA, snv.call=NA,tmp.id = paste(smp.grouping,probe.label,sep=".")) %>% 
    arrange(.,probe.label,smp.grouping)
  
  
  
  del.t <- .1
  max.iter <- 20
  
  g <- NULL
  if(visualize)
    g <- sapply(unique(snv.dat$probe.label),function(x)NA,simplify = F,USE.NAMES = T)
  
  for(probe.i in unique(snv.dat$probe.label)){
    
    resid.t <- NULL
    iter <- 1
    converge <- FALSE
    R2.adj <- NULL
    
    
    dat.pi <- snv.dat[snv.dat$probe.label == probe.i,]
    dat.pi$smp.grouping.EM <- relevel(as.factor(dat.pi$smp.grouping.EM),ref = "ref")
    
    while(iter < max.iter){
      frm <- frm0
      
      if(n_distinct(dat.pi$smp.grouping.EM)==1)
        frm <- update(frm,.~.-smp.grouping.EM)
      lmod.pi <- lm(frm, data = dat.pi)
      
      # We are only intrested when we have repeates/imperfect fit
      # and coocks distance is only defined for imperfect fit -> Floor the cooks distance
      cd <- cooks.distance(lmod.pi)
      cd[lmod.pi$residuals<1e-10] <- min(cd[1e-10 <lmod.pi$residuals ]) 
      
      outlier <- cd>(10/lmod.pi$df.residual) & scale(lmod.pi$fitted.values,center = T,scale = T)< -3
      if(any(outlier))
        lmod.pi <- lm(frm, data = dat.pi[!outlier,])
      
      R2.adj <- c(R2.adj,summary(lmod.pi)$adj.r.squared)
      
      # Fit a model that predicts the "ref" counterpart for each sample
      #-----------------------------------------------------------------
      dat.pi %>% select(.,c(nlCount, WT.nlCount.adj, adj.factor)) %>%
        mutate(smp.grouping.EM = "ref") %>% 
        predict.lm(lmod.pi,newdata = . ,se.fit = T,level = 1- call.rule$sig.level) -> prd.ref
      dof <- prd.ref$df
      # Note here we are using the predictive sd = sqrt(se^2 + resid.scale^2) to construct t.value and ci
      dat.pi %>% 
        mutate(.,
               ref.ft = prd.ref$fit,
               se = prd.ref$se.fit,
               resid.scale = prd.ref$residual.scale,
               resid.dof = dof,#prd.ref$df,
               t.value = ( (nlCount - bg.hat) - ref.ft)/sqrt(se^2 + resid.scale^2),
               ci.ref.lo = ref.ft - qt(1-(call.rule$sig.level)/2,df = resid.dof) * sqrt(se^2 + resid.scale^2),
               ci.ref.hi = ref.ft + qt(1-(call.rule$sig.level)/2,df = resid.dof) * sqrt(se^2 + resid.scale^2),
               in.ref.tmp = smp.grouping == "ref" | ci.ref.lo<(nlCount - bg.hat) & (nlCount - bg.hat)<ci.ref.hi & count< max(max.bg.count,mean(dat.pi$count[dat.pi$is.smp.ref=="Y"]))) %>% 
        mutate(., in.ref.tmp = c("N","Y")[1+in.ref.tmp]) -> dat.pi
      
      
      if(summary(lmod.pi)$adj.r.squared==max(R2.adj))
        best.soln <- dat.pi
      
      
      
      if(visualize){
        require(gridExtra)
        
        g1 <- ggplot(data = dat.pi,aes(x=adj.factor,y=(nlCount - bg.hat) - ref.ft) ) + 
          geom_point(aes(shape = is.smp.ref,colour = in.ref.tmp,size = WT.nlCount.adj)) +
          geom_smooth(method="lm",data= filter(dat.pi,in.ref.tmp == "Y")) +
          geom_text(data = filter(dat.pi,in.ref.tmp == "Y"),aes(label = count)) +
          geom_line(aes(x = adj.factor, y= ci.ref.lo-ref.ft), linetype = "dashed") +     
          geom_line(aes(x = adj.factor, y= ci.ref.hi-ref.ft), linetype = "dashed") + 
          geom_abline(slope = 0,intercept = 0,linetype = 3)+
          ggtitle(paste(probe.i,"R2.adj:",round(max(R2.adj),2)))
        
        
        g2 <- ggplot(data = dat.pi,aes(x=adj.factor,y=(nlCount - bg.hat)) ) + 
          geom_point(aes(shape = is.smp.ref,colour = in.ref.tmp,size = WT.nlCount.adj)) +
          geom_smooth(method="lm",data= filter(dat.pi,in.ref.tmp == "Y")) +
          geom_text(data = filter(dat.pi,in.ref.tmp == "Y"),aes(label = count)) +
          ggtitle(paste(probe.i,"R2.adj:",round(max(R2.adj),2)))
        
        g3 <- ggplot(data = dat.pi,aes(x=WT.nlCount.adj,y=(nlCount - bg.hat)) ) + 
          geom_point(aes(shape = is.smp.ref,colour = in.ref.tmp,size = WT.nlCount.adj)) +
          geom_smooth(method="lm",data= filter(dat.pi,in.ref.tmp == "Y")) +
          geom_text(data = filter(dat.pi,in.ref.tmp == "Y"),aes(label = count)) +
          ggtitle(paste(probe.i,"R2.adj:",round(max(R2.adj),2)))
        
        g4 <- ggplot(data.frame(iter=1:length(R2.adj),R2.adj),aes(x=iter,y=R2.adj))+geom_line()+geom_point()
        
        g[[probe.i]]  <- list(resid = g1, adj.fact = g2, WT = g3, R2 =g4)
        
        # print(do.call("grid.arrange",c(g[[probe.i]],ncol=2)))
      }
      
      
      
      
      #If no new points are picked up ==> Exact converagance 
      converge <- all(dat.pi$in.ref.tmp == dat.pi$in.ref)
      
      if(!converge & !is.null(resid.t))
        converge <- median(abs(resid.t - dat.pi$t.value)) < del.t 
      
      if(converge)
        break
      
      resid.t <- dat.pi$t.value
      dat.pi %>% mutate(.,smp.grouping.EM = smp.grouping) -> dat.pi
      dat.pi$smp.grouping.EM[dat.pi$in.ref.tmp=="Y"] <- "ref"
      dat.pi$smp.grouping.EM <- relevel(as.factor(dat.pi$smp.grouping.EM),ref = "ref")
      iter <- iter + 1
      
    } #END while
    
    #Update the fc.stats
    #-------------------
    pool.sd <- function(x){sqrt(sum(x^2))/length(x)}
    
    best.soln %>%
      select(.,c(smp.grouping, smp.grouping.EM,probe.label,nlCount,bg.hat,ref.ft,se,t.value,resid.dof,ci.ref.lo,ci.ref.hi,resid.dof,resid.scale)) %>% 
      group_by(.,smp.grouping) %>% summarise(.
                                             , probe.label=probe.label[1]
                                             , nlCount = mean(nlCount)
                                             , bg.hat = mean(bg.hat)
                                             , ref.ft = mean(ref.ft)
                                             , se = pool.sd(se)
                                             , resid.scale = resid.scale[1]
                                             , resid.dof = resid.dof[1] + length(unique(smp.grouping.EM[smp.grouping.EM!="ref"])) - 1
                                             , t.value = ( (nlCount - bg.hat) - ref.ft)/sqrt(se^2 + resid.scale^2)
                                             , ci.ref.lo = ref.ft - qt(1-(call.rule$sig.level)/2,df = resid.dof) * sqrt(se^2 + resid.scale^2)
                                             , ci.ref.hi = ref.ft + qt(1-(call.rule$sig.level)/2,df = resid.dof) * sqrt(se^2 + resid.scale^2)) %>% 
      transmute(.,
                estimate = (nlCount-bg.hat) - ref.ft,
                se = se,
                resid.scale = resid.scale,
                pred.sd =  sqrt(se^2 + resid.scale^2),
                t.value = t.value,
                p = pt(t.value,df = resid.dof,lower.tail = F),
                ci.lo = estimate + ci.ref.lo - ref.ft,
                ci.hi = estimate + ci.ref.hi - ref.ft,
                resid.dof = resid.dof,
                tmp.id = paste(smp.grouping,probe.i,sep = ".")) %>% filter(.,tmp.id %in% fc.stats$tmp.id) -> best.soln
    
    fc.stats[match(best.soln$tmp.id,fc.stats$tmp.id),colnames(best.soln)] <- best.soln
    
  }
  
  
  fc.stats$tmp.id <- NULL
  
  # Construct the snv.f2 array
  #---------------------------
  fc.stats %>% 
    gather(.,key = stat.type,value= value,-c(smp.grouping, probe.label)) %>% 
    reshape2:::acast(., smp.grouping ~ probe.label ~ stat.type) -> snv.f2
  
  snv.f2 <- snv.f2[,,c("count","estimate", "se","resid.scale","pred.sd", "t.value", "p","ci.lo", "ci.hi", "resid.dof","snv.call"), drop = F]
  
  
  # Make calls
  #-----------
  snv.calls <- make.snv.call(snv.f2 = snv.f2,call.rule = call.rule)
  snv.f2[,,"snv.call"] <- snv.calls$snv.call[dimnames(snv.f2[,,"snv.call", drop = F])[[1]], dimnames(snv.f2[,,"snv.call", drop = F])[[2]], drop = F ]
  
  
  # update snv.dat
  #---------------
  data.frame(smp.grouping = rownames(snv.calls$snv.call),snv.calls$snv.call,check.names = F,stringsAsFactors = F) %>%
    gather(.,key = probe.label,value = snv.call.EM,-smp.grouping) %>% left_join(x=snv.dat,y=.,by=c("smp.grouping","probe.label")) %>%
    mutate(.,snv.call = snv.call.EM) -> snv.dat
  
  
  return(list(snv.f2 = snv.f2, snv.dat = snv.dat, fc.stats = fc.stats, g = g))
}



#' This function performs post-estimation adjustment to remove possible bias due to sample normalization
#' @param snv.f2 the array result output containig the stats by probe and by sample needed for making calls 
#' @param visualize this logical if TRUE, returns the diagnostic plots for excuted function for debugging/evaluation purposes
#' @param use.likelihood.test this logical if TRUE, uses adjust the t.value using prediction sd as opposed to se
#'  
debias.by.sample <- function(snv.f2 = snv.f2, visualize = FALSE, use.likelihood.test){
  base.raw.bg <- 50
  
  sd.type <- "pred.sd"
  if(!use.likelihood.test)
    sd.type <- "se"
  
  smp.bias <- rep(NA,dim(snv.f2)[[1]])
  names(smp.bias) <- dimnames(snv.f2)[[1]]
  
  
  g <- NULL
  if(visualize)
    g <- sapply(dimnames(snv.f2)[[3]],function(x)NA,simplify = F,USE.NAMES = T)
  
  for(smp.i in dimnames(snv.f2)[[1]]){
    
    not.called.prbs <- dimnames(snv.f2)[[2]][snv.f2[smp.i,,"snv.call"] == 0] # NOTE we are using definite no's as opposed to undetected inclusive
    above.background.prbs <- dimnames(snv.f2)[[2]][snv.f2[smp.i,,"count"]> base.raw.bg]
    
    smp.bias[smp.i] <- 0
    if(!length(not.called.prbs)>0){
      warning(paste("For sample", smp.i, "debiasing could not be completed!\n"))
      next
    }
      
    
    if(length(intersect(not.called.prbs,above.background.prbs)) == 0){
      warning(paste("For sample", smp.i, "all no call probes were below",base.raw.bg,"debaising may be unreliable!\n"))
      smp.bias[smp.i] <- mean(snv.f2[smp.i, not.called.prbs,"estimate"])
    }else
      smp.bias[smp.i] <- mean(snv.f2[smp.i,intersect(not.called.prbs,above.background.prbs),"estimate"])
      
    

    if(visualize){
      #needs to be adjusted for multicategory calls!!!!!!!!!!
      # df <- data.frame(probe = dimnames(snv.f2)[[2]],
      #                  Est = snv.f2[smp.i,,"estimate"],
      #                  calls = c("NO","YES")[1+snv.f2[smp.i,,"snv.call"]],
      #                  above.bg.counts = (snv.f2[smp.i,,"count"]>base.raw.bg))
      # df$in.baseline <- c("NO","YES")[1+(df$above.bg.counts & df$calls == "NO")]
      # g[[smp.i]] <- ggplot(df,aes(x=probe,y=Est,colour = calls, shape = in.baseline, size = in.baseline)) + 
      #   geom_point(alpha = .75) + 
      #   geom_hline(yintercept = c(0,
      #                             mean(snv.f2[smp.i,intersect(not.called.prbs,above.background.prbs),"estimate"]),
      #                             median(snv.f2[smp.i,intersect(not.called.prbs,above.background.prbs),"estimate"])),linetype=1:3, colour = 1:3)+
      #   ggtitle(smp.i)+
      #   theme(axis.text.x = element_text(angle = 90, hjust = 1))      
    }
    
  }
  
  
  snv.f2[,,c("estimate","ci.hi","ci.lo")] <- sweep(snv.f2[,,c("estimate","ci.hi","ci.lo"), drop = F],MARGIN = 1,STATS = smp.bias,FUN = "-")
  snv.f2[,,"t.value"] <- abind:::adrop(snv.f2[,,"estimate",drop = F], drop = c(F,F,T))  /abind:::adrop(snv.f2[,,sd.type,drop = F], drop = c(F,F,T))    
  snv.f2[,,"p"] <- pt(q = abind:::adrop(snv.f2[,,"t.value", drop = F], drop = c(F,F,T)), df = abind:::adrop(snv.f2[,,"resid.dof", drop = F ], drop = c(F,F,T)),lower.tail = F)
  
  
  return(list(snv.f2=snv.f2,smp.bias = smp.bias, plots = g))
}


# Other convenience functions
#---------------------------------------

#' This function validates the the format of SNV.f2
#' @param snv output of eval.snv
#' @param sample.name The sample names to be verified. NULL makes is choose all sample names
#' @param sample.rename The sample renames to be verified. Null makes it equivalent to sample.name
#' @param return a list containing snv, sample.name and smaple.rename named as such

validate.snv <- function(snv, sample.name = NULL, sample.rename = NULL){
  
  snv.f2 <- snv$snv.f2
  if(length(dimnames(snv.f2))!=3)
    stop("snv.f2 needs to be named for each of its 3 dimensions")
  
  if(any(dimnames(snv.f2)[[3]] !=  c("count", "estimate", "se","resid.scale","pred.sd", "t.value", "p","ci.lo", "ci.hi", "resid.dof","snv.call")))
    stop("dimnames(snv.f2)[[3]] is expected to be count estimate, ci.hi, ci.lo, p, snv.call, se, resid.scale, pred.sd, t.value, resid.dof and snv.call")
  
  if(is.null(sample.name))
    sample.name <- dimnames(snv.f2)[[1]]
  
  if(any(!sample.name %in% dimnames(snv.f2)[[1]])){
    warning("sample.name(s) needs to be in the dimnames(snv.f2)[[1]]. They are ignored")
    sample.name <- dimnames(snv.f2)[[1]]
  }
  
  
  if(max(table(sample.name))>1)
    stop("sample.names should be unique")
  
  if(is.null(sample.rename)){
    default.names.map <- snv$snv.dat %>% distinct(smp.grouping,.keep_all = T) %>% select(smp.label,smp.grouping)
    rownames(default.names.map) <- default.names.map$smp.grouping
    sample.rename <- default.names.map[sample.name,"smp.label"]
    names(sample.rename) <- sample.name
  }
  
  
  if(length(sample.rename) != length(sample.name)){
    warning("sample.rename length needs to match the length of sample.names to be ploted. They are ignored")
    sample.rename <- sample.name
    names(sample.rename) <- sample.name
  }
  
  if(any(duplicated(dimnames(snv.f2)[[2]])))
    stop("Multiple probes for the same variant loci is not supported")
  
  return(list(snv = snv, sample.name = sample.name, sample.rename = sample.rename))
}



#' This is a convenience function to convert the snv.f2 element of snv object into a long data.frame
#' @param snv output of eval.snv
#' @param sample.name The sample names to be verified. NULL makes is choose all sample names
#' @param sample.rename The sample renames to be verified. Null makes it equivalent to sample.name
#' @return a data.frame of all samples and the corresponding statistics computed

snvf2.to.data.frame <- function(snv, sample.name = NULL, sample.rename = NULL){
  
  # Validate input
  #---------------
  tmp <- validate.snv(snv = snv, sample.name = sample.name, sample.rename = sample.rename)
  snv <- tmp$snv
  sample.name <- tmp$sample.name
  sample.rename <- tmp$sample.rename
  
  snv.f2 <- snv$snv.f2
  smps <- 1:dim(snv.f2)[[1]]
  names(smps) <- dimnames(snv.f2)[[1]] 
  
  snv.f2.df <- sapply(smps, FUN = function(i) data.frame(smp.grouping = dimnames(snv.f2)[[1]][i],
                                                        probe.label = rownames(snv.f2[i,,]),
                                                        snv.f2[i,,], stringsAsFactors = F), simplify = F, USE.NAMES = T) %>% bind_rows()
    
  snv.f2.df <- mutate(.data = snv.f2.df,
                      snv.call = factor(snv$call.rule$categories$call.type[1+(snv.f2.df$snv.call)],levels = snv$call.rule$categories$call.type),
                      smp.label = sample.rename[snv.f2.df$smp.grouping])
  
  attr(snv.f2.df, "sample.name") <- sample.name
  attr(snv.f2.df, "sample.rename") <- sample.rename
  attr(snv.f2.df, "sample.order") <- smps
  
  return(snv.f2.df)
}


#' This is a convenience function to extract call summary table from snv object
#' @param snv output of eval.snv
#' @param sample.name The sample names to be verified. NULL makes is choose all sample names
#' @param sample.rename The sample renames to be verified. Null makes it equivalent to sample.name
#' @param mat.format either long or wide character string. It determins whether the output should be in long wor wide format 
#' @return a data.frame of calls for each sample x probe combo

get.snv.call.summary.mat <- function(snv, sample.name = NULL, sample.rename = NULL, mat.format = c("long","wide")){
  
  mat.format <- match.arg(mat.format,  c("long","wide"),several.ok = F)
  snvf2.df <- snvf2.to.data.frame(snv = snv, sample.name = sample.name, sample.rename = sample.rename)
  
  call.df <- snvf2.df %>% select(smp.label, probe.label, snv.call)
  if(mat.format == "long")
    return(call.df)
  
  call.df <- call.df %>%
    spread(.,key = probe.label, value = snv.call) %>%
    `rownames<-` (.$smp.label) %>%
    dplyr:::select(-smp.label)
  
  return(call.df)
}