#' This function runs the workflow from top to bottom

run.analysis <- function(){
  
  
  warnings.paragraph <- ""
  
  #======================================================
  #           Load/Install required libraries
  #======================================================
  
  resultlib <- suppressWarnings(try(runloadlibraries(), silent = TRUE))
  warnings()
  if(class(resultlib) == "try-error")
  {
    cat("LOG:could not load libraries",file=log,sep='\n',append=TRUE)
    cat(resultlib,file=log,sep='\n',append=TRUE)
    cat("document.write('<p>Could not load libraries</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
    stop("could not load libraries")
  } else{
    print("Finished loading R libraries")
    cat("LOG:Finished loading R libraries",file=log,sep='\n\n',append=TRUE)
    cat("document.write('<p>Finished loading R libraries</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE) 
  }
  
  
  #======================================================
  #               Load and verify data
  #======================================================
  
  # FUTURE DESIGN REV: consider extracting "matrix set" information outside this step and at the time of use within the modules
  #                    for example, probe.set.matrix and KEGG.set.matrix etc. could be gleaned probe annotation by calling the 
  #                    proper functionality within load_and_clean.R
  
  annot.obj <- NULL
  annot.obj <- tryCatch(expr = construct.data.and.annotations(path.data = path.data,
                                                              path.sampleannot = path.sampleannot,
                                                              path.probeannot = path.probe.annot,
                                                              snvarg.ref.ID = snvarg.ref.ID,
                                                              prb.annotation.id.column = probe.annotation.id.column,
                                                              
                                                              #prb.annotation.id.column = gene.annotation.id.column,
                                                              prb.annotation.KEGG.id.column = probe.annotation.KEGG.id.column,
                                                              prb.annotation.KEGG.Gene.Name = probe.annotation.KEGG.Gene.Name,
                                                              prb.annotation.celltype.id.column = probe.annotation.celltype.id.column,
                                                              prb.set.column = probe.set.column, # <------------------------------ comment this out to see the behavior when no probeset info is avaialable
                                                              min.prb.set.size = min.prb.set.size,
                                                              
                                                              #min.prb.set.size = min.gene.covariate,
                                                              min.KEGG.set.size = min.KEGG.set.size,
                                                              sample.annotation.id.column = sample.annotation.id.column,
                                                              sampleannot.variables = sampleannot.variables,
                                                              sampleannot.variabletypes = sampleannot.variabletypes,
                                                              sampleannot.referencelevels = sampleannot.referencelevels,
                                                              exclude.probes = NULL,
                                                              exclude.samples = NULL,
                                                              draw.color.legend = T,
                                                              min.samples = 2,
                                                              plottypearg = plottypearg,
                                                              log = log)
                        ,error = function(er){
                          ertmp <- paste("Data load and preprocess is aborted: the following error(s) occurred\n",er)
                          cat("document.write('<p> Data load and preprocess is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                          class(ertmp) <- "error"
                          return(ertmp)
                        }
                        ,finally = {})
  
  
  if(class(annot.obj) == "error" )
  {
    print("Failed in upload and preprocess")
    cat("LOG: Failed in upload and preprocess",file=log,sep='\n',append=TRUE)
    cat(annot.obj,file=log,sep='\n',append=TRUE)
    
    warnings.paragraph <- paste(warnings.paragraph,annot.obj,"\n")
    stop(warnings.paragraph)
  }else{
    # Assign to local variables
    #--------------------------
    rawdata <- annot.obj$rawdata
    processedData <- annot.obj$processed
    annot <- annot.obj$annot
    pannot <- annot.obj$pannot
    # run.batch.id.column = !colnames(annot)!="run.batch.id"
    heatmap.annot <- annot.obj$heatmap.annot#[,!run.batch.id.column,drop=F]
    annotcols <- annot.obj$annotcols#[,!run.batch.id.column,drop=F]
    annotcols2 <- annot.obj$annotcols2#[!run.batch.id.column,drop=F]
    sampleannot.variables <- annot.obj$sampleannot.variables#[!run.batch.id.column]
    sampleannot.variabletypes <- annot.obj$sampleannot.variabletypes
    sampleannot.referencelevels <- annot.obj$sampleannot.referencelevels
    prb.set.matrix <- annot.obj$prb.set.matrix
    celltype.set.matrix <- annot.obj$celltype.set.matrix
    warnings.paragraph <- annot.obj$warnings.paragraph
    KEGG.set.matrix <- annot.obj$KEGG.set.matrix
    prb.set.matrix.QC <- prb.set.matrix  # <-----------------------------------what's difference between these?
    
    
    #post hoc elimination of irrelvant warnings: 
    #This (along with other warning messaging) 
    #is to be handled by better design in next rev!
    #----------------------------------------------
    
    if(!celltypes.arg.run)
      warnings.paragraph <- gsub("Warning: No celltype probe annotation is provided\n","",warnings.paragraph)
    
    if(!pathviewmodule.arg.run)
      warnings.paragraph <- gsub("Warning: No KEGG id annotation is provided\n","",warnings.paragraph)
    
  }
  
  
  #======================================================
  #                       Fusion              
  #======================================================
  fusionarg.run <- FALSE
  fusion.dir <- NULL
  fusion.analyzed <- NULL
  if("fusion" %in% names(attributes(rawdata)))
    fusionarg.run <- TRUE
  
  if(fusionarg.run){
    
    cat("running fusion analysis module",file=log,sep='\n\n',append=TRUE)
    cat(timestamp(),file=log,sep='\n',append=TRUE) 
    
    print("Starting fusion analysis")
    cat("LOG:Starting fusion analysis",file=log,sep='\n\n',append=TRUE)
    cat("document.write('<p>Starting fusion analysis</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
    
    # Ascribe the values
    #------------------
    fusion.raw <- attr(rawdata,"fusion")
    fusion.pannot <- attr(pannot,"fusion")
    ctrl.raw <- attr(rawdata,"ctrl")
    ctrl.pannot <- attr(pannot,"ctrl")
    
    # Add housekeepers to fusion.raw
    HK.ids <- ctrl.pannot$ProbeID[tolower(ctrl.pannot$Control.Type) == tolower("Housekeeping")]
    
    if(all(rownames(fusion.raw) %in% rownames(ctrl.raw))){
      fusion.raw <- cbind(fusion.raw,ctrl.raw[rownames(fusion.raw),HK.ids,drop=F])
      fusion.pannot <- rbind(fusion.pannot, ctrl.pannot[HK.ids,])
      
      fusion.dat <- tryCatch(expr = {readFusionData(f.data = fusion.raw,
                                                    f.pannot = fusion.pannot, 
                                                    junctionRef = NULL,
                                                    hk.QC.threshold = 30,#this is the minimum count for any HK probes, otherwise, QC failure
                                                    pvalue.ep.ttest = fusionarg.pvalue.endprobe,#this is the p value for t test of end probes
                                                    pvalue.junc.test = fusionarg.pvalue.junction,
                                                    pvalue.ep.lmtest = 0.0001, #this is the p value for lm test of end probes
                                                    junc.pvalue.threshold = 0.0001, #this is the pvalue threshold for fusion calling from outlier test
                                                    fusion.min.ref.cluster = 10, #this is the minimum sample size for reference dataset to run cluster analysis
                                                    fusion.min.sample.size = 20, #this is the minimum sample size of final dataset to do cluster analysis
                                                    fusion.pvalue.threshold.less.probes = 0.005,#this is  the p value threshold for genes with less than 3 junction probes, such as NTRK1
                                                    sd.from.water.lane = 1.174,
                                                    end.cluster.arg = FALSE,#turn off clustering on imbalance probes
                                                    QC.excluded = TRUE, #this is the control whether or not to exclude QC failed sample from analysis
                                                    ref.sample.imbalanced = NA, #this specifies the reference sample which will be used for linear regression
                                                    myhk = HK.ids,
                                                    logfile =log)}
                             , error = function(er){
                               ertmp <- paste("Fusion data upload failed. Fusion analysis is aborted: the following error(s) occurred\n",er)
                               cat("document.write('<p> Fusion data upload failed. Fusion analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                               class(ertmp) <- "error"
                               return(ertmp)
                             }
                             , finally = {})
      
      if(!"error" %in% class(fusion.dat)){
        
        tmp <- tryCatch(expr = getFusionCall(fusion.dat)
                        ,error = function(er){
                          ertmp <- paste("Fusion analysis is aborted: the following error(s) occurred\n",er)
                          cat("document.write('<p> Fusion analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                          class(ertmp) <- "error"
                          return(ertmp)
                        }
                        , finally = {})
        
        if(!"error" %in% class(tmp))
          fusion.analyzed <- tmp
        
        
        if(!is.null(fusion.analyzed)){
          print("Generating fusion plots")
          cat("LOG:Generating fusion plots",file=log,sep='\n\n',append=TRUE)
          cat("document.write('<p>Generating fusion plots</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
          
          fusion.dir <- tryCatch(expr = {plot.fusion(fusion.dat = fusion.dat,
                                                     fusion.analyzed = fusion.analyzed,
                                                     type = c(c("call.summary","heatmap","by.gene"),"QC"[HK.ids]),
                                                     file.extension = c(plottypearg,"csv"),
                                                     path.to.fusion.results = path.to.fusion.results,
                                                     fusion.pannot = fusion.pannot,
                                                     fusionarg.pvalue.endprobe = fusionarg.pvalue.endprobe)}
                                 , error = function(er){
                                   ertmp <- paste("Fusion plotting failed. Fusion analysis is aborted: the following error(s) occurred\n",er)
                                   cat("document.write('<p> Fusion analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                                   class(ertmp) <- "error"
                                   return(ertmp)
                                 }
                                 , finally = {})
          
          if("error" %in% class(fusion.dir)){
            fusion.dir <- NULL
          }else{
            print("Fusion analysis completed")
            cat("LOG:Fusion analysis completed",file=log,sep='\n\n',append=TRUE)
            cat("document.write('<p>Fusion analysis completed</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
          }
            
        }else{
          fusionarg.run <- FALSE
          warnings.paragraph <- paste(warnings.paragraph,"Error:",fusion.dir,"\n")
        }
      }else{
        fusionarg.run <- FALSE
        warnings.paragraph <- paste(warnings.paragraph,"Error:",fusion.dat,"\n")
      }
      
    }else{
      fusionarg.run <- FALSE
    }
    
  }
  
  
  #======================================================
  #                       SNV              
  #======================================================
  snvarg.run <- FALSE
  SNV.dir <- NULL
  snv.analyzed <- NULL
  if(all(c("snv.ref","snv") %in% names(attributes(rawdata))))
    snvarg.run <- TRUE
  
  if(snvarg.run){
    
    cat("running SNV analysis module",file=log,sep='\n\n',append=TRUE)
    cat(timestamp(),file=log,sep='\n',append=TRUE) 
    
    print("Starting SNV analysis")
    cat("LOG:Starting SNV analysis",file=log,sep='\n\n',append=TRUE)
    cat("document.write('<p>Starting SNV analysis</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
    
    # Ascribe the values
    #------------------
    snv.raw <- sapply(c("snv.ref","snv") ,FUN = function(x) attr(rawdata,x),simplify = F,USE.NAMES = T)
    if(all(colnames(snv.raw$snv.ref)  %in% colnames(snv.raw$snv)) & ncol(snv.raw$snv.ref) == ncol(snv.raw$snv))
      snv.raw <- rbind(snv.raw$snv.ref[,colnames(snv.raw$snv)],snv.raw$snv)
    
    snv.sannot <- plyr::rbind.fill(attr(annot,"snv.ref"),attr(annot,"snv"))
    rownames(snv.sannot) <- c(rownames(attr(annot,"snv.ref")), rownames(attr(annot,"snv")))
    snv.sannot <- snv.sannot[rownames(snv.raw),sapply(snv.sannot,function(x) !any(is.na(x))),drop=F]
    # snv.sannot <- rbind(attr(annot,"snv.ref")[,"run.batch.id",drop=F],attr(annot,"snv")[,"run.batch.id",drop=F])[rownames(snv.raw),,drop=F]
    
    # Pre-process the input data
    #---------------------------
    tmp <- tryCatch(expr = {preprocess.snv(raw.dat = snv.raw,
                                           sannot = snv.sannot,
                                           pannot = attr(pannot,"snv"),
                                           loci.to.estimate = NULL,
                                           smp.names = rownames(attr(rawdata,"snv")),
                                           samps.ref = rownames(attr(rawdata,"snv.ref")),
                                           adj.frm =  as.formula("~ (SNVPOS_A + SNVPOS_D)/(SNVPOS_B + SNVPOS_E)"),
                                           smp.relabels = NULL,
                                           analysis.batch = "cartridge.id")}
                    ,error = function(er){
                      ertmp <- paste("SNV analysis is aborted: the following error(s) occurred\n",er)
                      cat("document.write('<p> SNV analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                      class(ertmp) <- "error"
                      return(ertmp)
                    }
                    , finally = {})
    
    if(!"error" %in% class(tmp))
      snv.dat <- tmp
    
    # Perform the SNV analysis
    #-------------------------
    snv.call.rule <- make.call.rule(fc.breaks = c(-Inf, snvarg.foldchangethreshold.hi, snvarg.foldchangethreshold) ,
                                call.type = c("Undetected","No call","Detected"),
                                sig.level = c(1, snvarg.pvaluethreshold.hi, snvarg.pvaluethreshold),
                                raw.background = 50)
    
    tmp <- tryCatch(expr ={ eval.snv(snv.dat = snv.dat,
                                     call.rule = snv.call.rule,
                                     subtract.background = T,
                                     post.est.EM = snvarg.em,
                                     post.est.debias = snvarg.debias)}
                    , error = function(er){
                      ertmp <- paste("SNV analysis is aborted: the following error(s) occurred\n",er)
                      cat("document.write('<p> SNV analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                      class(ertmp) <- "error"
                      return(ertmp)
                    }
                    ,finally = {})
    
    if(!"error" %in% class(tmp)){
      print("Generating SNV plots")
      cat("LOG:Generating SNV plots",file=log,sep='\n\n',append=TRUE)
      cat("document.write('<p>Generating SNV plots</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      
      snv.analyzed <- tmp
      
      SNV.dir <- list()
      # Generate and store plots
      #-------------------------
      SNV.dir <- plot.snv(snv = snv.analyzed,
                          sample.name = NULL ,
                          path = path.to.SNV.results ,
                          sample.rename = NULL,
                          file.extension = c(plottypearg,"csv"),
                          type = c("summary","by.sample","QC"))
      
      print("SNV analysis completed")
      cat("LOG:SNV analysis completed",file=log,sep='\n\n',append=TRUE)
      cat("document.write('<p>SNV analysis completed</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      
    }else{
      snvarg.run <- FALSE
      SNV.dir <- NULL
    }
    
    
  }
  # remove the run.batch.id cloumn for downstream analysis
  annot <- annot.obj$annot <- annot[,!colnames(annot) %in% c("run.batch.id","Batch.ID"), drop=F]
  arg.env$annot <- annot
  
  #======================================================
  #                   Normalization                
  #======================================================
  
  # This section could use some clean up and transferring the logic into DE source
  
  rawData.list  <- processedData.list <- normData.list <- 
    sapply(X = unique(pannot$Analyte.Type),simplify = F,USE.NAMES = T, FUN = function(x) NA)
  normData.list.2 <- list()  
  
  cat("running normalization module",file=log,sep='\n',append=TRUE)
  cat(timestamp(),file=log,sep='\n',append=TRUE)
  for(l in unique(pannot$Analyte.Type)){
    if(l %in% c("FUSION","SNV"))
      next
    
    pannot.sub <- pannot[(pannot$Analyte.Type) == l,]
    intersect.analyte <- intersect(dimnames(rawdata)[[2]],dimnames(pannot.sub)[[1]])
    
    processedData.list[[l]] <- processedData[,intersect.analyte]
    rawData.list[[l]]  <- rawdata[,intersect.analyte]
    
    if(! l %in% c("mRNA","protein"))
      stop(paste("No method for normalizing analyte type",l,"is available"))
    
    if(l == "mRNA"){
      
      ERCCs <- pannot.sub[pannot.sub$Control.Type %in% c("Negative","Positive"),probe.annotation.id.column]
      HKs <- pannot.sub[pannot.sub$Control.Type =="Housekeeping",probe.annotation.id.column]
      
      #some default values set for the case where we don't normalize (check why we need to pass data into normalization func if don't want normalization)
      codeset.HKs <- pannot.sub[!rownames(pannot.sub) %in% ERCCs,probe.annotation.id.column]
      codeset.HKs <- codeset.HKs[0:min(10,length(codeset.HKs))]
      if(length(HKs)>0)
        codeset.HKs <- HKs
      normmodule.arg.method <- "none"
      n.HKs <- NA
      auto.HKs <- FALSE
      
      
      norm.opts <- tryCatch(expr = {get.mRNA.normalization.opt(raw.dat = as.matrix(rawData.list[[l]]),
                                                               normalize = "mRNA" %in% normmodule.analytes.run,
                                                               normmodule.norm.probes = normmodule.norm.probes,
                                                               normmodule.nchoos = normmodule.nchoos,
                                                               probe.annotation.id.column = probe.annotation.id.column,
                                                               pannot = pannot)},
                            error = function(er) return(er),
                            finally ={})
      
      
      if(!class(norm.opts) == "try-error"){
        n.HKs <- norm.opts$n.HKs
        auto.HKs <- norm.opts$auto.HKs
        codeset.HKs <- norm.opts$codeset.HKs
        normmodule.arg.method <- norm.opts$HK.selection.method 
      }else{
        warnings.paragraph <- paste(warnings.paragraph,"\n","Failed to find housekeeping mRNA probes. Disregard any Housekeeping content plot in normalization module\n",norm.opts)
      }
      
      
      #Force exlusion of ERCCs from data to be normalized
      tmp <- as.matrix(rawData.list[[l]])
      tmp.raw <- tmp[,!colnames(tmp) %in% ERCCs,drop=F]
      
      tmp <- processedData.list[[l]]
      tmp.processed <- tmp[,!colnames(tmp) %in% ERCCs,drop=F]
      normresult <- list()
      if(ncol(tmp.raw)>0 & ncol(tmp.processed)>0){
        
        normresult <- normalizerawdata(raw=tmp.raw,
                                       processed = tmp.processed,
                                       data.type = data.type,
                                       method=normmodule.arg.method,
                                       choose.method = "geNorm",
                                       n.HKs = n.HKs,
                                       auto.HKs = auto.HKs,
                                       codeset.HKs = codeset.HKs,
                                       path.to.normalization.results=path.to.normalization.results,
                                       log=log,
                                       plottypearg=plottypearg,
                                       path.results=path.results,
                                       path.inc=path.inc,
                                       prb.annots = pannot)
      }else{
        warnings.paragraph <- paste(warnings.paragraph,"Warning: No informative mRNA probes is left in this dataset","\n")
      }
      #if(all.equal(normresult$norm.factors,rep(0,length(normresult$norm.factors)))) {
      if (data.type == "normalized") {
        # scaling.factors = log2(rowMeans(tmp.raw)/rowMeans(tmp.processed))
        # add or subtract norm.factors?
        scaling.factors = log2((tmp.raw+1)/(tmp.processed+1)) + normresult$norm.factors
        scaling.factors = sweep(scaling.factors,colMeans(scaling.factors),MARGIN=2,FUN="-")
        normresult$scaling.factors = scaling.factors
      } else {
        normresult$scaling.factors = normresult$norm.factors
      }
    }
    if(l == "protein"){
      
      #Set the default to Processed data unless processed is different from raw and user explicitly asked
      #to use raw. The default is overwritten unless the user turn off the normalization
      normresult <- list()
      normresult$norm.factors <- rep(0,nrow(rawData.list[[l]]))
      normresult$normalized.data <- log2(processedData.list[[l]]+1)
      
      if(!isTRUE(all.equal(rawData.list[[l]],processedData.list[[l]],check.attributes=F)) & data.type == "raw")
        normresult$normalized.data <- log2(rawData.list[[l]]+1)
      
      if("protein" %in% normmodule.analytes.run){
        
        # Manual option: User selected probes for normalization 
        #----------------------------------------------------------------------------------------------
        norm.ref.prot.names  <- get.probes.for.analyte (probes = normmodule.norm.probes,
                                                        probe.annotation.id.column = probe.annotation.id.column,
                                                        pannot = pannot,
                                                        analyte.type = "protein")
        
        neg.colnames <- as.character(pannot[which(pannot$Control.Type == "Protein_NEG"),probe.annotation.id.column])
        
        normresult <- preprocess.prot(raw.dat = rawData.list[[l]],
                                      neg.colnames = neg.colnames,
                                      norm.method = "geomean.stable1", 
                                      norm.ref.prot.names = norm.ref.prot.names,
                                      normmodule.nchoos = normmodule.nchoos,
                                      smp.annot = NULL,
                                      prb.annots = pannot,
                                      path.to.normalization.results = path.to.normalization.results,
                                      log = log,
                                      plottypearg = plottypearg,
                                      path.results = path.results,
                                      path.inc = path.inc)
      } else{
        warning("Warning: No protein normalization will be performed - assuming the input data is already normalized")
        cat("LOG: Warning: No protein normalization will be performed - assuming the input data is already normalized",file=log,sep='\n\n',append=TRUE)
        cat("document.write('<p>Warning: No protein normalization will be performed - assuming the input data is already normalized</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
        
      }
      
      if (data.type == "normalized") {
        scaling.factors = log2((rawData.list[[l]]+1)/(processedData.list[[l]]+1)) + normresult$norm.factors
        scaling.factors = sweep(scaling.factors,colMeans(scaling.factors),MARGIN=2,FUN="-")
        normresult$scaling.factors = scaling.factors
      } else {
        normresult$scaling.factors = normresult$norm.factors
      } 
    }
    
    
    normData.list[[l]]      <- normresult$normalized.data
    normData.list.2[[l]] <- normresult
    
    #Change the global value of normmodule analyte and nchoose
    if(ncol(normresult$normalized.data)==0 | all(normresult$norm.factors == 0)){
      normmodule.analytes.run <<- normmodule.analytes.run[normmodule.analytes.run!=l] #arg.env$normmodule.analytes.run
      normmodule.nchoos <<-  normmodule.nchoos[names(normmodule.nchoos)!=l] #arg.env$normmodule.nchoos <-
    }
    
    # If no data is left eliminate the analyte type from dataset
    if(ncol(normresult$normalized.data)==0)
      normData.list[[l]] <- normData.list.2[[l]] <- NULL
  }
  
  # CHECK FOR NULL
  if ("mRNA" %in% names(rawData.list) & "mRNA" %in% names(normData.list.2)) {
    mrna.posCtrls <- pannot[pannot$Control.Type =="Positive",probe.annotation.id.column]
    mrna.geomeanPOS=exp(apply(log(rawData.list[["mRNA"]][,mrna.posCtrls]),1,mean))
    mrna.meangeoPOS=mean(mrna.geomeanPOS)
    mrna.pos.factors= mrna.meangeoPOS/mrna.geomeanPOS 
    mrna.norm.factors = 2^normData.list.2$mRNA$scaling.factors
    if(is.vector(mrna.norm.factors)) {
      if(is.null(names(mrna.norm.factors))) names(mrna.norm.factors) = rownames(rawData.list[["mRNA"]])
      mrna.norm.factors = mrna.norm.factors[match(rownames(rawData.list[["mRNA"]]),names(mrna.norm.factors))]
    }
    if(is.matrix(mrna.norm.factors)) {
      if(is.null(rownames(mrna.norm.factors))) rownames(mrna.norm.factors) = rownames(rawData.list[["mRNA"]])
      mrna.norm.factors = mrna.norm.factors[match(rownames(rawData.list[["mRNA"]]),rownames(mrna.norm.factors)),]
    }
    #heatmap.annot = cbind(heatmap.annot, mrna.norm.factor = mrna.norm.factors, mrna.pos.factor = mrna.pos.factors)
  }
  if ("protein" %in% names(rawData.list) & "protein" %in% names(normData.list.2)) {
    protein.norm.factors = 2^(normData.list.2$protein$scaling.factors)
    if(is.vector(protein.norm.factors)) {
      if(is.null(names(protein.norm.factors))) names(protein.norm.factors) = rownames(rawData.list[["protein"]])
      protein.norm.factors = protein.norm.factors[match(rownames(rawData.list[["protein"]]),names(protein.norm.factors))]
    }
    if(is.matrix(protein.norm.factors)) {
      if(is.null(rownames(protein.norm.factors))) rownames(protein.norm.factors) = rownames(rawData.list[["protein"]])
      protein.norm.factors = protein.norm.factors[match(rownames(rawData.list[["protein"]]),rownames(protein.norm.factors)),]
    }

    #heatmap.annot = cbind(heatmap.annot, protein.norm.factor = protein.norm.factors)
  }
  
  #heatmap.cols = heatmap.annot
  heatmap.cols <- annot.to.aes(annot = heatmap.annot,reverse.columns=F)
  names(heatmap.cols[["annotcols2"]]) = colnames(heatmap.annot)
  # !!make annotcols2 consistent with heatmap.cols
  #for (annot.name in names(annotcols2)) {
  #  heatmap.cols$annotcols[,annot.name] <- annotcols[,annot.name]
  #  heatmap.cols$annotcols2[[annot.name]] <- annotcols2[[annot.name]] 
  #}
  #for(name in names(heatmap.cols[["annotcols2"]])) {
  #  if(!is.null(heatmap.cols[["annotcols2"]][[name]]) & is.numeric(heatmap.annot[,name])) {
  #    names(heatmap.cols[["annotcols2"]][[name]])[-2] = paste(names(heatmap.cols[["annotcols2"]][[name]])[-2],round(range(heatmap.annot[,name]),2))
  #  }
  #}
  # merge normalized data
  #----------------------
  cbind.all <- function(x, y){
    rowMatch <- match(rownames(x), rownames(y))
    cbind(x, y[rowMatch,])
  }
  
  e <- Reduce(cbind.all, normData.list)
  
  
  if(!is.null(e)){
    # Remove controls that are not being used
    #----------------------------------------
    control.probe.ids <- pannot[pannot$Is.Control == 1,probe.annotation.id.column]
    if(length(which(colnames(e)%in%control.probe.ids))>0)
      e <- e[,-which(colnames(e)%in%control.probe.ids)]
    
    # relabel e for report
    #---------------------
    e.relabeled <- e
    colnames(e.relabeled) <- paste(pannot[colnames(e),"Probe.Label"],pannot[colnames(e),"Analyte.Type"],sep="-")
    #write.csv(e.relabeled, file = paste(path.to.normalization.results,"//ALL_normalized log2 data.csv",sep=""))
    
    log.temp.df3 <- data.frame("log" = c("ALL - normalized log2 count data"))
    write.table(log.temp.df3,file=paste(path.to.normalization.results,"//ALL_normalized_data.csv",sep=""), col.names = F, row.names = F)
    write.table(e.relabeled,file=paste(path.to.normalization.results,"//ALL_normalized_data.csv",sep=""), append = T, sep = ",", col.names = NA, row.names = T)
    
    linear.temp.df3 <- data.frame("linear" = c("", "ALL - normalized linear count data"))
    write.table(linear.temp.df3,file=paste(path.to.normalization.results,"//ALL_normalized_data.csv",sep=""), append = T, sep = ",", col.names = F, row.names = F)
    write.table(2^e.relabeled,file=paste(path.to.normalization.results,"//ALL_normalized_data.csv",sep=""), append = T, sep = ",", col.names = NA, row.names = T)
    
  }else{
    QCmodule.arg.run <- 
      DEmodule.arg.run <- 
      gsamodule.arg.run <- 
      pathviewmodule.arg.run <- 
      celltypes.arg.run <- pathscores.arg.run <- RPDarg.run  <- SPDarg.select.probe.analysis.run<- FALSE
  }
  
  
  
  #======================================================
  #            Overview Descriptive Analyses
  #======================================================
  QCresult <- NULL
  if(QCmodule.arg.run)
  {
    suppressWarnings(write.table(head(as.data.frame(annot)),file=log,sep='\t',append=TRUE))
    cat("running QC module",file=log,sep='\n',append=TRUE)
    #HTML NOTE - the HTML for the QC is embedded in the other R-file
    cat(timestamp(),file=log,sep='\n',append=TRUE)
    
    QCresult <- tryCatch(expr = {runQC(normalized=e, 
                                       rawdata=rawdata,
                                       snv.analyzed = snv.analyzed,
                                       fusion.analyzed = fusion.analyzed,
                                       thresholding.value=thresholding.value,
                                       probe.annot=pannot,
                                       rawDataAll.list=rawData.list,
                                       normalizedData.list=normData.list,
                                       HKnormalized = normData.list.2$"mRNA"$HKnormalized,
                                       HKs= normData.list.2$"mRNA"$HKs,
                                       covariates= as.data.frame(annot),
                                       covariates.type = sampleannot.variabletypes,
                                       prb.sets=prb.set.matrix.QC,
                                       annotcols=annotcols,
                                       annotcols2=annotcols2,
                                       codecols=codecols2,
                                       path.to.QC.results=path.to.QC.results,
                                       log=log,
                                       plottypearg=plottypearg,
                                       path.results=path.results,
                                       path.inc=path.inc,
                                       path.to.csvs=path.to.csvs,
                                       heatmap.cols = heatmap.cols)}
                         , error = function(er){
                           ertmp <- paste("Overview is aborted: the following error(s) occurred\n",er)
                           #cat(ertmp)
                           cat("document.write('<p> Overview aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                           class(ertmp) <- "error"
                           return(ertmp)
                         }
                         , finally = {})
    
    if(class(QCresult) == "error" )
    {
      print("Failed to run Overview")
      cat("LOG: Failed to run Overview",file=log,sep='\n',append=TRUE)
      cat(QCresult,file=log,sep='\n',append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,QCresult,"\n")
      QCmodule.arg.run <- FALSE
      QCresult <- NULL
    }
    
  }
  
  
  #======================================================
  #                DE Analysis                  
  #======================================================
  
  
  if(DEmodule.arg.run){
    source("R_source/NSDE.r")
    source("R_source/DE_func.r")
    mrna.bg.mean = mrna.bg.dispersion = protein.bg.mean = protein.bg.dispersion = NULL
    mrna.rawdata = rawData.list[["mRNA"]]
    if(!is.null(mrna.rawdata) & !is.null(normData.list.2$mRNA)) {
      #mrna.norm.factors = 2^(normData.list.2$mRNA$scaling.factors)
      mrna.posCtrls <- pannot[pannot$Control.Type =="Positive",probe.annotation.id.column]
      mrna.pos.raw <- mrna.rawdata[,colnames(mrna.rawdata) %in% mrna.posCtrls,drop=F]
      mrna.negCtrls <- pannot[pannot$Control.Type =="Negative",probe.annotation.id.column]
      mrna.neg.raw <- mrna.rawdata[,colnames(mrna.rawdata) %in% mrna.negCtrls,drop=F]
      mrna.bg.mean = rowMeans(mrna.neg.raw,na.rm=T)
      if(ncol(mrna.neg.raw)==0)
        mrna.bg.mean[1:length(mrna.bg.mean)] = 1
      mrna.bg.dispersion = max(median((apply(mrna.neg.raw,2,var) - colMeans(mrna.neg.raw))/colMeans(mrna.neg.raw)^2),MIN.PHI)
      if(is.null(mrna.bg.dispersion) || is.na(mrna.bg.dispersion))
        mrna.bg.dispersion <- MIN.PHI
    }
    protein.rawdata = rawData.list[["protein"]]
    if(!is.null(protein.rawdata) & !is.null(normData.list.2$protein)) {
      #protein.norm.factors = 2^(normData.list.2$protein$norm.factor)
      #if(is.null(names(protein.norm.factors))) names(protein.norm.factors) = rownames(protein.rawdata)
      protein.negCtrls <- pannot[pannot$Control.Type =="Protein_NEG",probe.annotation.id.column]
      protein.neg.raw <- protein.rawdata[,colnames(protein.rawdata) %in% protein.negCtrls,drop=F]
      protein.bg.mean = rowMeans(protein.neg.raw,na.rm=T)
      if(ncol(protein.neg.raw)==0)
        protein.bg.mean[1:length(protein.bg.mean)] = 1
      protein.bg.dispersion = max(median((apply(protein.neg.raw,2,var)-colMeans(protein.neg.raw))/colMeans(protein.neg.raw)^2),MIN.PHI)
      if(is.null(protein.bg.dispersion) || is.na(protein.bg.dispersion))
        protein.bg.dispersion = MIN.PHI
    }
    
    
    if(is.null(mrna.rawdata) && is.null(protein.rawdata) ){
      print("Failed to run DE Analysis")
      cat("LOG: Failed to run DE Analysis",file=log,sep='\n',append=TRUE)
      cat("document.write('<p>DE aborted: No valid probe was selected for DE</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,"No valid probe was selected for DE. DE was aborted","\n")
      DEresults <- NULL
      DEmodule.arg.run=FALSE
    }
    
    DEresults <- NULL
    DEmodule.arg.predictors = reconcile.variables.with.annot(annot,variables=DEmodule.arg.predictors,
                                                             variabletypes=NULL,
                                                             referencelevels=NULL)$variables
    categorical.predictors = DEmodule.arg.predictors[sampleannot.variabletypes[DEmodule.arg.predictors] == 'categorical'] 
    for (predictor in categorical.predictors) {
      # If all groups do not have at least 3 elements, remove from DE analysis
      group.counts = as.data.frame(table(annot[predictor]))
      if(sum(group.counts$Freq >= 3) < nrow(group.counts)) {
        DEmodule.arg.predictors = setdiff(DEmodule.arg.predictors, predictor)
        warnings.paragraph <- paste(warnings.paragraph, "Differential Expression was not completed due to low group size (less than three) for: ", predictor, "\n")
      }
    }
    if(length(DEmodule.arg.predictors)==0 & isTRUE(DEmodule.arg.run)){
      print("Failed to run DE Analysis")
      cat("LOG: Failed to run DE Analysis",file=log,sep='\n',append=TRUE)
      cat("document.write('<p>DE aborted: No valid predictor was selected for DE</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,"No valid predictor was selected for DE. DE was aborted","\n")
      DEresults <- NULL
      DEmodule.arg.run=FALSE
    }
  }
  
  if(DEmodule.arg.run)
  {
    #gene.set.matrix = pannot[,gene.set.columns]; dimnames(gene.set.matrix)[[1]] = dimnames(pannot)[[1]]
    
    cat("running DE module",file=log,sep='\n\n',append=TRUE)
    cat(timestamp(),file=log,sep='\n',append=TRUE)
    
    DEmodule.arg.confounders = reconcile.variables.with.annot(annot,variables=DEmodule.arg.confounders,
                                                              variabletypes=NULL,
                                                              referencelevels=NULL)$variables
    
    
    
    cat("predictors:",file=log,sep='\n',append=TRUE)
    suppressMessages(suppressWarnings(write.table(DEmodule.arg.predictors,file=log,sep='\t',append=TRUE)))
    cat("confounders:",file=log,sep='\n',append=TRUE)
    suppressMessages(suppressWarnings(write.table(DEmodule.arg.confounders,file=log,sep='\t',append=TRUE)))
    cat("predictor var types:",file=log,sep='\n',append=TRUE)
    suppressMessages(suppressWarnings(write.table(sampleannot.variabletypes[DEmodule.arg.predictors],file=log,sep='\t',append=TRUE)))
    cat("confounder var types:",file=log,sep='\n',append=TRUE)
    suppressMessages(suppressWarnings(write.table(sampleannot.variabletypes[DEmodule.arg.confounders],file=log,sep='\t',append=TRUE)))
    
    # Perform filtering out of low expressors and constant
    #-----------------------------------------------------
    
    
    filtered <- filter.out.low.counts(raw = rawdata,prb.annots=pannot,thresholding.frequency = thresholding.frequency, thresholding.value = thresholding.value,thresholding.perform = thresholding.perform)
    warnings.paragraph <- paste(warnings.paragraph,filtered$warnings,"\n")
    filtered.raw <- rawdata[,!filtered$prune[colnames(rawdata)],drop=FALSE]
    const.prbs <- colnames(rawdata)[apply(rawdata,2,sd)==0]
    expression.data = filtered.raw[,!colnames(filtered.raw)%in%c(const.prbs,control.probe.ids)]
    #filtered.raw <- filtered.raw[,!colnames(filtered.raw)%in%const.prbs]
    #expression.data = filtered.raw[,!colnames(filtered.raw)%in%control.probe.ids]
    filtered.data.list = list()
    if(!is.null(mrna.rawdata) & !is.null(normData.list$mRNA)) 
      filtered.data.list[["mRNA"]] <- expression.data[,colnames(expression.data)%in%colnames(rawData.list[["mRNA"]]),drop=FALSE]
    if(!is.null(protein.rawdata) & !is.null(normData.list$protein)) 
      filtered.data.list[["protein"]] <- expression.data[,colnames(expression.data)%in%colnames(rawData.list[["protein"]]),drop=FALSE]
    
    prb.sets = prb.set.matrix[colnames(expression.data),,drop=FALSE]
    
    if ("mRNA" %in% names(filtered.data.list)) {
      if(is.matrix(mrna.norm.factors)) {
        mrna.norm.factors = mrna.norm.factors[,colnames(filtered.data.list$mRNA)]
      }
    }
    if ("protein" %in% names(filtered.data.list)) {
      if(is.matrix(protein.norm.factors)) {
        protein.norm.factors = protein.norm.factors[,colnames(filtered.data.list$protein)]
      }
    }
    if(length(const.prbs)>0)
      warnings.paragraph <- paste(warnings.paragraph,paste(const.prbs,collapse = ","),"dropped from DE for being constant","\n")
    
    
    DEresults = tryCatch(expr = {run.DE.analyses(filtered.data.list=filtered.data.list,
                                                 predictors=DEmodule.arg.predictors,
                                                 confounders=DEmodule.arg.confounders,
                                                 annot=annot,
                                                 predictor.var.type=sampleannot.variabletypes[DEmodule.arg.predictors],
                                                 confounder.var.type=sampleannot.variabletypes[DEmodule.arg.confounders],
                                                 sampleannot.referencelevels=sampleannot.referencelevels,
                                                 pval.adjustment = DEmodule.arg.pval.adjustment,
                                                 prb.sets=prb.sets,  
                                                 path.to.DE.results=path.to.DE.results,
                                                 log=log,
                                                 plottypearg=plottypearg,
                                                 path.results=path.results,
                                                 path.inc=path.inc,
                                                 path.to.csvs=path.to.csvs,
                                                 probe.annotation.id.column = probe.annotation.id.column,
                                                 prb.annots = pannot,
                                                 mrna.norm.factors = mrna.norm.factors,
                                                 protein.norm.factors = protein.norm.factors,
                                                 mrna.bg.mean = mrna.bg.mean,
                                                 protein.bg.mean = protein.bg.mean,
                                                 mrna.bg.dispersion = mrna.bg.dispersion,
                                                 protein.bg.dispersion = protein.bg.dispersion,
                                                 optimzation = DEmodule.arg.method)},
                         error = function(er){
                           ertmp <- paste("DE is aborted: the following error(s) occurred in execution of DE\n",er)
                           #cat(ertmp)
                           cat("document.write('<p>DE aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                           class(ertmp) <- "DE.error"
                           return(ertmp)
                         },
                         finally = {})
    
    
    
    if(class(DEresults) == "DE.error" )
    {
      print("Failed to run DE Analysis")
      cat("LOG: Failed to run DE Analysis",file=log,sep='\n',append=TRUE)
      cat(DEresults,file=log,sep='\n',append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,DEresults,"\n")
      DEmodule.arg.run <- FALSE
      DEresults <- NULL
      
    }else{
      ## add gene sets corresponding to the DE hits:
      if(length(DEresults$sig.gene.sets)>0)
      {
        for(i in 1:length(DEresults$sig.gene.sets))
        {
          tempcolumn = 1*is.element(rownames(prb.set.matrix),DEresults$sig.gene.sets[[i]])
          prb.set.matrix.QC = cbind(prb.set.matrix.QC,tempcolumn)
          colnames(prb.set.matrix.QC)[ncol(prb.set.matrix.QC)]=names(DEresults$sig.gene.sets)[i]
        }
      }
      
    }
    
  }
  
  
  
  #======================================================
  #               Gene set Analysis
  #======================================================
  
  if(gsamodule.arg.run){
    wtmp <- NULL
    if(!DEmodule.arg.run){
      wtmp <- paste("Geneset analysis will be not be conducted because the required DE analysis was not performed\n")
      cat("document.write('<p>Geneset analysis will be not be conducted because the required DE analysis was not performed</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      gsamodule.arg.run <- FALSE
    }
    if(is.null(prb.set.matrix)){
      wtmp <- paste("Geneset analysis will be not be conducted because no geneset annotation is provided\n")
      cat("document.write('<p>Geneset analysis will be not be conducted because no geneset annotation is provided</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      gsamodule.arg.run <- FALSE
    }
    if(!is.null(wtmp)){
      warning(wtmp)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",wtmp,"\n")
    }
    rm(wtmp) 
  }
  
  
  if(gsamodule.arg.run)
  {
    cat("running gene set analysis module",file=log,sep='\n\n',append=TRUE)
    cat(timestamp(),file=log,sep='\n',append=TRUE)  
    
    
    gsaresults <- tryCatch(expr = run.gene.set.analyses(DEresults,prb.sets=prb.sets,
                                                        #prb.sets=prb.set.matrix[colnames(e[,!filtered$prune[colnames(e)],drop=FALSE]),,drop=FALSE],  #,
                                                        prb.set.names=colnames(prb.set.matrix),
                                                        pval.adjustment=DEmodule.arg.pval.adjustment,
                                                        log=log,plottypearg=plottypearg,path.results=path.results,path.inc=path.inc,
                                                        path.to.csvs=path.to.csvs,path.to.GSA.results=path.to.GSA.results,prb.annots=pannot)
                           , error=function(er){
                             ertmp <- paste("Geneset analysis is aborted: the following error(s) occurred\n",er)
                             #cat(ertmp)
                             cat("document.write('<p>Geneset Analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                             class(ertmp) <- "error"
                             return(ertmp)
                           }
                           , finally = {})
    
    if(class(gsaresults) == "error" )
    {
      print("Failed to run Geneset Analysis")
      cat("LOG: Failed to run Geneset Analysis",file=log,sep='\n',append=TRUE)
      cat(gsaresults,file=log,sep='\n',append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,gsaresults,"\n")
      gsamodule.arg.run <- FALSE
      gsaresults <- NULL
    }
    
    
  }
  
  
  #======================================================
  #                 Pathview Plots
  #======================================================
  # each node represents a protein family; do the proteins in the assay need to be incorporated somehow?
  # each node can correspond to multiple genes; in which case it is an average of fold-changes or t-statistics
  # if incorporating the proteins, do the gene.names need to be corrected (official gene symbols)
  
  if(pathviewmodule.arg.run){
    
    if(is.null(KEGG.set.matrix)){
      pathviewmodule.arg.run=FALSE
      warning("Was unable to validate KEGG annoation - Pathview will not be run")
      cat("LOG: Warning: Was unable to validate KEGG annoation - Pathview will not be run",file=log,sep='\n',append=TRUE)
      cat("document.write('<p>Warning: Was unable to validate KEGG annoation - Pathview will not be run</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      paste(warnings.paragraph,"Warning: Was unable to validate KEGG annoation - Pathview will not be run","\n")
    }
    
    if(!DEmodule.arg.run){
      pathviewmodule.arg.run=FALSE
      warning("DE was not completed - Pathview will not be run")
      cat("LOG: Warning: DE was not completed - Pathview will not be run",file=log,sep='\n',append=TRUE)
      cat("document.write('<p>Warning: DE was not completed - Pathview will not be run</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      paste(warnings.paragraph,"Warning:DE was not completed - Pathview will not be run","\n")
    }
    
    
    if(length(pathviewmodule.arg.keggIDs)== 0){
      pathviewmodule.arg.run=FALSE
      warning("No pathway IDs were provided for pathview display")
      cat("LOG: Warning: no valid KEGG ID selected",file=log,sep='\n',append=TRUE)
      cat("document.write('<p>Warning: No valid KEGG ID is selected - Pathview will not be run</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      paste(warnings.paragraph,"Warning:No valid KEGG ID is selected - Pathview will not be run","\n")
    }
    
    if(!testCon()){
      pathviewmodule.arg.run=FALSE
      warning("Internet connection failure - Pathview will not be run")
      cat("LOG: Warning: Internet connection failure - Pathview will not be run",file=log,sep='\n',append=TRUE)
      cat("document.write('<p>Warning: Internet connection failure - Pathview will not be run</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      paste(warnings.paragraph,"arning: Internet connection failure - Pathview will not be run","\n")
    }
  }
  
  
  if(pathviewmodule.arg.run)
  {
    cat("running Pathview module",file=log,sep='\n\n',append=TRUE)
    cat(timestamp(),file=log,sep='\n',append=TRUE)
    # clean the custom IDs to make them parse right: rule: 5 digits, all numbers.
    #<----------------------------------
    
    pathview.results <- tryCatch(expr = run.pathview(DEresult=DEresults, 
                                                     prb.annot=pannot,
                                                     keggIDs=pathviewmodule.arg.keggIDs,
                                                     keggNames=pathviewmodule.arg.keggNames,
                                                     colorscheme = pathviewmodule.arg.view.colorscheme,
                                                     pvalthresh=pathviewmodule.arg.pvalthres,
                                                     view=c("KEGG"),
                                                     predictors=DEmodule.arg.predictors,
                                                     path.to.pathview.results=path.to.pathview.results,
                                                     log=log,
                                                     path.results=path.results,
                                                     path.inc=path.inc,
                                                     path.to.csvs=path.to.csvs)
                                 , error = function(er){
                                   ertmp <- paste("Pathview Analysis is aborted: the following error(s) occurred\n",er)
                                   #cat(ertmp)
                                   cat("document.write('<p> Pathview Analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                                   class(ertmp) <- "error"
                                   return(ertmp)
                                 }
                                 , finally = {})
    
    
    setwd(analysis.path)
    if(class(pathview.results) == "error"){
      print("Failed to run pathview")
      cat("LOG: Failed to run pathview",file=log,sep='\n',append=TRUE)
      cat(pathview.results,file=log,sep='\n',append=TRUE)
      pathviewmodule.arg.run <- FALSE
      pathview.results <- NULL
    }
    
    #note pathview result right not is just a warning
    if(length(pathview.results)>0){
      cat(pathview.results[[1]])
      #warnings.paragraph <- paste(warnings.paragraph, pathview.results, "\n")
    }
    
  }
  
  
  #======================================================
  #                   Cell type Analysis
  #======================================================
  celltypescoreslist <- NULL
  if(celltypes.arg.run)  
  {
    tempcelltype = replace(pannot[,probe.annotation.celltype.id.column],pannot$Cell.Type=="",NA)
    names(tempcelltype) = dimnames(pannot)[[1]]
    
    celltypecontrasts = read.csv(celltypes.arg.contrastsfile,row.names=1)
    
    celltypes.arg.covariates = reconcile.variables.with.annot(annot,variables=celltypes.arg.covariates,
                                                              variabletypes=NULL,
                                                              referencelevels=NULL)$variables
    
    filtered.2 <- filter.out.low.counts(raw = rawdata, prb.annots = pannot, thresholding.frequency = thresholding.frequency, 
                                        thresholding.value = thresholding.value, thresholding.perform = thresholding.perform)
    warnings.paragraph <- paste(warnings.paragraph, filtered.2$warnings, "\n")
    
    e.filtered <- e[,!filtered.2$prune[colnames(e)], drop = FALSE]
    const.prbs <- colnames(rawdata)[apply(rawdata, 2, sd) == 0]
    e.filtered <- e.filtered[,!colnames(e.filtered) %in% const.prbs]
    
    if(length(const.prbs) > 0)
      warnings.paragraph <- paste(warnings.paragraph, paste(const.prbs, collapse = ","),"dropped from Cell Type Profiling for being constant","\n")
    
    celltyperesult <- tryCatch(expr = immune.cell.abundance.analysis(data = e.filtered,
                                                                     prb.annot = pannot,
                                                                     celltype = tempcelltype,
                                                                     use.all.genes=(celltypes.arg.probeselection!="DynamicallySelect"),
                                                                     lowcor=.2,
                                                                     cell.type.pval.thresh=celltypes.arg.pvalthreshold,
                                                                     celltypes.arg.showraw=celltypes.arg.showraw,
                                                                     celltypes.arg.showrelative=celltypes.arg.showrelative,
                                                                     celltypecontrasts=celltypecontrasts,
                                                                     plot.against.covariates=length(celltypes.arg.covariates)>0,
                                                                     covariates=annot[,celltypes.arg.covariates,drop=FALSE],
                                                                     covariate.type=sampleannot.variabletypes[celltypes.arg.covariates],
                                                                     #annotcols=as.matrix(annotcols[,celltypes.arg.covariates]),
                                                                     annotcols=annotcols[,celltypes.arg.covariates,drop=FALSE],
                                                                     annotcols2=annotcols2[match(celltypes.arg.covariates,dimnames(annot)[[2]])],
                                                                     log=log,
                                                                     plottypearg=plottypearg,
                                                                     path.results=path.results,
                                                                     path.inc=path.inc,
                                                                     path.to.csvs=path.to.csvs,
                                                                     path.to.celltype.results=path.to.celltype.results)
                               , error=function(er){
                                 ertmp <- paste("cell type scoring is aborted: the following error(s) occurred\n",er)
                                 #cat(ertmp)
                                 cat("document.write('<p> Cell Type Analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                                 class(ertmp) <- "error"
                                 return(ertmp)
                               }
                               , finally = {})
    
    if(class(celltyperesult) == "error"){
      print("Failed to run Cell Type analysis")
      cat("LOG: Failed to run Cell Type analysis",file=log,sep='\n',append=TRUE)
      cat(celltyperesult,file=log,sep='\n',append=TRUE)
      
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",celltyperesult,"\n")
      celltypes.arg.run <- FALSE
      celltyperesult <- NULL
    }else{
      celltypescoreslist = celltyperesult$celltypescoreslist
      warnings.paragraph <- paste(warnings.paragraph, celltyperesult$CTP.warnings.paragraph)
    }
    
  }
  
  
  #======================================================
  #                   Pathway Scoring Analysis
  #======================================================
  pathscore.results <- NULL
  if(pathscores.arg.run){
    wtmp <- character(0)
    if(is.null(prb.set.matrix)){
      wtmp0 <- paste("Pathway scoring can not be completed as there are no valid probesets defined")
      cat(paste("document.write('<p>",wtmp0,"</p>');"), file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      wtmp <- paste(wtmp,wtmp0,"\n")
      pathscores.arg.run <- FALSE
    }
    
    pathscores.arg.covariates <- make.names(pathscores.arg.covariates)
    pathscores.arg.covariates <- pathscores.arg.covariates[pathscores.arg.covariates %in% colnames(annot)]
    if(length(pathscores.arg.covariates)==0){
      wtmp0 <- paste("Pathway scoring can not be completed as there are no valid covariates selected")
      cat(paste("document.write('<p>",wtmp0,"</p>');"), file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      wtmp <- paste(wtmp,wtmp0,"\n")
      pathscores.arg.run <- FALSE
    }
    
    
    if(!is.null(pathscores.arg.adjust.for)){
      if(any(!pathscores.arg.adjust.for %in% colnames(annot))){
        wtmp0 <- paste("Pathway scoring can not be completed due to invalid adjustment covariate:",paste(pathscores.arg.adjust.for,collapse = ","))
        cat(paste("document.write('<p>,",wtmp0, ",</p>');"), file=paste(path.inc,"//status.js",sep=""),append=TRUE)
        wtmp <- paste(wtmp,wtmp0,"\n")
        pathscores.arg.run <- FALSE
      }
    }
    
    
    if(length(wtmp)>0){
      warning(wtmp)
      cat(wtmp,file=log,sep='\n',append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",wtmp,"\n")
    }
    rm(wtmp) 
  }
  
  if(pathscores.arg.run)  
  {
    # run the pathway scoring function:
    
    # data = normalized data matrix (no distinguishing between analyze types)
    # gene.sets.list: named list of gene set vectors for PC1 scoring
    # adjust.for: data frame of covariates to adjust for. ONLY used under PC1 <--------------------
    # signaturematrix: matrix of gene weights for predefined signatures
    # requireallgenes: T/F for whether to use incomplete gene signatures (ie those with some genes not in codeset)
    # mingenes: min number of genes for a geneset or signature to be calculated
    # prunedprobes: only removed from PC1 scoring
    
    
    if(!is.null(pathscores.arg.adjust.for))
      pathscores.arg.adjust.for <- annot[,pathscores.arg.adjust.for,drop=FALSE]
    
    pathscores.arg.covariates.type <- NA
    if(!is.null(pathscores.arg.covariates))
      pathscores.arg.covariates.type <- sampleannot.variabletypes[pathscores.arg.covariates]
    
    filtered.3 <- filter.out.low.counts(raw = rawdata, prb.annots = pannot, thresholding.frequency = thresholding.frequency, 
                                        thresholding.value = thresholding.value, thresholding.perform = thresholding.perform)
    warnings.paragraph <- paste(warnings.paragraph, filtered.3$warnings, "\n")
    
    e.filtered <- e[,!filtered.3$prune[colnames(e)], drop = FALSE]
    const.prbs <- colnames(rawdata)[apply(rawdata, 2, sd) == 0]
    e.filtered <- e.filtered[,!colnames(e.filtered) %in% const.prbs]
    
    if(length(const.prbs) > 0)
      warnings.paragraph <- paste(warnings.paragraph, paste(const.prbs, collapse = ","),"dropped from Pathway Scoring for being constant","\n")
    
    pathscore.results <- tryCatch(expr = run.pathway.scoring(data=e.filtered,
                                                             path.scoring.method=pathscores.arg.method,
                                                             #PC1 arguments:
                                                             prb.set.matrix=prb.set.matrix,
                                                             mingenes.pc1=5,
                                                             prunedprobes=NULL,
                                                             adjust.for=pathscores.arg.adjust.for, 
                                                             # existing scores arguments
                                                             signaturematrix.predefined=NULL,
                                                             requireallgenes=FALSE,
                                                             mingenes.existing=5, 
                                                             # other arguments:
                                                             plot.against.covariates=length(pathscores.arg.covariates)>0,
                                                             covariates=as.data.frame(annot[,pathscores.arg.covariates,drop=F]),
                                                             covariate.type=pathscores.arg.covariates.type,
                                                             annotcols=annotcols,
                                                             annotcols2=annotcols2,
                                                             log=log,
                                                             plottypearg=plottypearg,
                                                             path.results=path.results,
                                                             path.inc=path.inc,
                                                             path.to.csvs=path.to.csvs,
                                                             path.to.pathscore.results=path.to.pathscore.results)
                                  ,error = function(er){
                                    ertmp <- paste("Pahtway scoring is aborted: the following error(s) occurred\n",er)
                                    #cat(ertmp)
                                    cat("document.write('<p> Pathway Scoring is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                                    class(ertmp) <- "error"
                                  }
                                  ,finally ={} )
  }
  
  if(pathscores.arg.run){
    
    if(class(pathscore.results) == "error"){
      print("Failed to run Pathway Scoring Analysis")
      cat("LOG: Failed to run Pahtway Scoring Analysis",file=log,sep='\n',append=TRUE)
      cat("document.write('<p> Failed to run Pathway Scoring Analysis due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
      cat(pathscore.results,file=log,sep='\n',append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",pathscore.results,"\n")
      pathscore.results  <- NULL
      pathscores.arg.run <- FALSE
    } else{
      warnings.pathscore <- pathscore.results[[2]]
      warnings.paragraph <- paste(warnings.paragraph, warnings.pathscore, "\n")
    }
  }
  
  
  #======================================================
  #                   SPD Analysis
  #======================================================
  SPD.dir <- NULL
  setwd(analysis.path)
  
  if(SPDarg.select.probe.analysis.run){
    
    wtmp <- validate.SPD.arguments(dat = e,smp.annots = annot)
    if(length(wtmp)>0)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",wtmp)
    
    
    if(length(SPDarg.probe.names.to.describe)>0){
      SPD.dir <- tryCatch(expr = {run.SPD.module(dat = e,
                                                 prb.annots = pannot,
                                                 smp.annots = annot,
                                                 color.cat.annots = annotcols2,
                                                 probe.ids.to.describe = SPDarg.probe.names.to.describe,
                                                 describe.probes.by = SPDarg.describe.probes.by,
                                                 stratify.probe.desription.by = NULL,
                                                 trend.by = SPDarg.trend.expression.by,
                                                 order.trend.by = SPDarg.order.expression.trend.by,
                                                 stratify.trend.by = SPDarg.stratify.expression.trend.by,
                                                 generate.interactnet.plot= SPDarg.interactnet.run,
                                                 interactnet.adjustfor = SPDarg.interactnet.adjustfor,
                                                 file.extension = plottypearg,
                                                 path.to.SPD.results = path.to.SPD.results)}
                          , error = function(er){
                            ertmp <- paste("Probe Descriptive Analysis is aborted: the following error(s) occurred\n",er)
                            #cat(ertmp)
                            cat("document.write('<p> Probe Descriptive Analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                            class(ertmp) <- "SPD.error"
                            return(ertmp)
                          }
                          , finally ={})
      # Generate error message if error was caught
      #-------------------------------------------
      if(class(SPD.dir) == "SPD.error"){
        print("Failed in Probe Descriptive Analysis")
        cat("LOG: Failed in Probe Descriptive Analysis",file=log,sep='\n',append=TRUE)
        cat(SPD.dir,file=log,sep='\n',append=TRUE)
        warnings.paragraph <- paste(warnings.paragraph,"Warning:",SPD.dir,"\n")
        SPD.dir <- NULL
        # SPDarg.select.probe.analysis.run <- FALSE
      }
      
    }else{
      wtmp <- "No probes remain in the set of probes to describe. Therefore, SPD analysis will not be completed"
      warning(wtmp)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",wtmp,"\n")
    }
    
    if(is.null(SPD.dir))
      SPDarg.select.probe.analysis.run <- FALSE
    
  }# End of SPD
  
  
  #======================================================
  #                   RPD Analysis
  #======================================================
  setwd(analysis.path)
  
  RPD.dir <- NULL
  
  if(RPDarg.run){
    wtmp <- validate.RPD.arguments(dat = e,prb.annots = pannot,smp.annots = annot)
    if(length(wtmp)>0)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",wtmp)
    
    
    
    RPD.dir <- tryCatch(expr = {run.RPD(dat = e, 
                                        prb.annots = pannot,
                                        smp.annots = annot,
                                        prb.grp.members = RPDarg.probe.pairs.to.describe,
                                        var.to.describe.probes.by = RPDarg.describe.probes.by,
                                        var.to.trend.by = RPDarg.trend.expression.by,
                                        var.to.order.obs.by = RPDarg.order.expression.trend.by,
                                        var.to.stratify.trend.by = RPDarg.stratify.expression.trend.by,
                                        color.cat.annots = annotcols2,
                                        file.extension = plottypearg,
                                        path.to.RPD.results = path.to.RPD.results)}
                        , error = function(er){
                          ertmp <- paste("Related Analytes Analysis is aborted: the following error(s) occurred\n",er)
                          #cat(ertmp)
                          cat("document.write('<p> Related Analytes Analysis is aborted due to error</p>');", file=paste(path.inc,"//status.js",sep=""),append=TRUE)
                          class(ertmp) <- "SPD.error"
                          return(ertmp)
                        }
                        , finally ={})
    # Generate error message if error was caught
    #-------------------------------------------
    if(class(RPD.dir) == "SPD.error"){
      print("Failed in Related Analytes Analysis")
      cat("LOG: Failed in Related Analytes Analysis",file=log,sep='\n',append=TRUE)
      cat(RPD.dir,file=log,sep='\n',append=TRUE)
      warnings.paragraph <- paste(warnings.paragraph,"Warning:",RPD.dir,"\n")
      RPD.dir <- NULL
    }
    
    if(is.null(RPD.dir))
      RPDarg.run <- FALSE
    
  }# End of RPD
  
  
  #======================================================
  #       Copy results into user-specified direcotry
  #======================================================
  cat("making results file directories",file=log,sep='\n\n',append=TRUE)
  
  # append the date and time to the analysis name
  # for the purpose of copying files:
  #----------------------------------------------
  advanced.analysis.name.for.file.copy <- paste(advanced.analysis.name,gsub(":","-",substr(Sys.time(),1,16)))
  results.basedir <- paste(path.results.copy,"//",advanced.analysis.name.for.file.copy,sep="")
  
  
  
  # make directories:
  #------------------
  dir.create(results.basedir)
  results.copydir <- paste(results.basedir, "//", "results", sep="")
  
  dir.create(results.copydir)
  
  # and copy results:
  #------------------
  cat("copying results files",file=log,sep='\n\n',append=TRUE)
  setwd(path.results)
  tfiles = dir()[-grep(pattern = "log|csvs",dir())]
  for(j in 1:length(tfiles))
  {
    invisible(file.copy(from = tfiles[j],to = results.copydir ,recursive = T))
  }
  
  # copy the resources
  results.copydir = paste(results.basedir, "//", "resources", sep="")
  dir.create(results.copydir)
  
  setwd(path.resources)
  tfiles = dir()
  for(j in 1:length(tfiles))
  {
    invisible(file.copy(from = tfiles[j],to = results.copydir ,recursive = T))
  }
  
  setwd(analysis.path)
  
  
  
  #======================================================
  #     Bundle the output of analysis into a list
  #======================================================
  
  #set some default values
  if(!exists("normresult"))
    normresult <- NULL
  if(!exists("annot.obj"))
    annot.obj <- NULL
  if(!exists("QCresult"))
    QCresult <- NULL
  if(!exists("DEresults"))
    DEresults<- NULL
  if(!exists("gsaresults"))
    gsaresults<- NULL
  if(!exists("pathview.results"))
    pathview.results<- NULL
  if(!exists("celltyperesult"))
    celltyperesult<- NULL
  if(!exists("celltypescoreslist"))
    celltypescoreslist<- NULL
  if(!exists("pathscore.results"))
    pathscore.results <- NULL
  if(!exists("SNV.dir"))
    SNV.dir <- NULL
  if(!exists("snv.call.rule"))
    snv.call.rule <- NULL
  if(!exists("fusion.dir"))
    fusion.dir <- NULL
  
  # return all analysis results:
  analysisresults = list(annot.obj =  annot.obj,
                         normresult = normData.list.2$mRNA,   # mRNA only
                         e = e,
                         QCresult = QCresult,
                         DEresults = DEresults,
                         gsaresults = gsaresults,
                         pathview.results = pathview.results,
                         celltyperesult = celltyperesult,
                         celltypescoreslist = celltypescoreslist,
                         QCmodule.arg.run = QCmodule.arg.run,
                         normmodule.arg.run=normmodule.arg.run,
                         DEmodule.arg.run=DEmodule.arg.run,
                         DEmodule.arg.predictors=DEmodule.arg.predictors,
                         DEmodule.arg.confounders=DEmodule.arg.confounders,
                         gsamodule.arg.run=gsamodule.arg.run,
                         pathviewmodule.arg.run=pathviewmodule.arg.run,
                         #pathviewmodule.arg.keggnames=pathviewmodule.arg.keggNames,
                         #pathviewmodule.arg.keggnames=as.character(pathview.results[[3]]),
                         #keggIDs=pathviewmodule.arg.keggIDs,
                         #keggIDs=as.character(pathview.results[[2]]),
                         pathview.display.order.list = pathview.results[[2]],
                         celltypes.arg.run = celltypes.arg.run,
                         pathscores.arg.run = pathscores.arg.run,
                         SPD.dir=SPD.dir,
                         SPDarg.select.probe.analysis.run = SPDarg.select.probe.analysis.run,
                         RPD.dir=RPD.dir,
                         RPDarg.run = RPDarg.run,
                         SNV.dir = SNV.dir,
                         SNV.call.rule = snv.call.rule,
                         snvarg.run = snvarg.run,
                         fusion.dir = fusion.dir,
                         fusionarg.run = fusionarg.run,
                         warnings.paragraph=warnings.paragraph,
                         celltypes.arg.covariates=celltypes.arg.covariates,
                         pathscores.arg.covariates=pathscores.arg.covariates,
                         probe.set.matrix.QC=prb.set.matrix.QC,
                         signatures.trim = pathscore.results[[1]],
                         arg.env = arg.env,
                         prb.annots = pannot,
                         results.basedir = results.basedir)
  
  return(analysisresults)
}

